"""
The System resource is a collection of miscellaneous server-wide
resources such as server properties, server directories, the
configuration store, Web Adaptors, and licenses.
"""
from __future__ import absolute_import
from __future__ import print_function
from .._common import BaseServer

########################################################################
class SystemManager(BaseServer):
    """
    The System resource is a collection of miscellaneous server-wide
    resources such as server properties, server directories, the
    configuration store, Web Adaptors, and licenses.
    """
    _json = None
    _json_dict = None
    _con = None
    _url = None
    _resources = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis,
                 initialize=False):
        """
        Constructor

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        url                    Required string. The machine URL.
        ------------------     --------------------------------------------------------------------
        gis                    Optional string. The GIS or Server object.
        ------------------     --------------------------------------------------------------------
        initialize             Optional string. Denotes whether to load the machine properties at
                               creation (True). Default is False.
        ==================     ====================================================================

        """

        super(SystemManager, self).__init__(gis=gis,
                                     url=url)
        self._con = gis
        if url.lower().endswith("/system"):
            self._url = url
        else:
            self._url = url + "/system"
        if initialize:
            self._init(gis)
    #----------------------------------------------------------------------
    def __str__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    @property
    def server_properties(self):
        """
        Gets the server properties for the site as an object.
        """
        return ServerProperties(url=self._url + "/properties",
                                connection=self._con,
                                initialize=True)
    #----------------------------------------------------------------------
    @property
    def _directories(self):
        """
        Gets the server directory object as a list.
        """
        directs = []
        url = self._url + "/directories"
        params = {
            "f" : "json"
        }
        res = self._con.get(path=url,
                            params=params)
        for direct in res['directories']:
            directs.append(
                ServerDirectory(url=url + "/%s" % direct["name"],
                                connection=self._con,
                                initialize=True))
        return directs
    #----------------------------------------------------------------------
    def directories(self):
        """
        :returns:
            The server directory object in a list.
        """
        return DirectoryManager(system=self)
    #----------------------------------------------------------------------
    def _get_directory(self, name):
        """
        Retrieves a single directory registered with ArcGIS Server.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        name                   Required string. The name of the registered directory.
        ==================     ====================================================================

        :returns:
            The ArcGIS Server directory as an object.

        """
        url = self._url + "/directories"
        params = {
            "f" : "json"
        }
        res = self._con.get(path=url,
                            params=params)
        for direct in res['directories']:
            if name.lower() == direct['name'].lower():
                return ServerDirectory(url=url + "/%s" % direct["name"],
                                       connection=self._con,
                                       initialize=True)
        return None
    #----------------------------------------------------------------------
    def _register(self,
                 name,
                 physical_path,
                 directory_type,
                 max_age,
                 cleanup_mode="NONE",
                 description=None):
        """
        Registers a new server directory. While registering the server
        directory, you can also specify the directory's cleanup parameters.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        name                   Required string. The name of the server directory.
        ------------------     --------------------------------------------------------------------
        physical_path          Required string. The absolute physical path of the server directory.
        ------------------     --------------------------------------------------------------------
        directory_type         Required string. The type of server directory.
        ------------------     --------------------------------------------------------------------
        max_age                Required integer. The length of time a file in the directory needs
                               to be kept before it is deleted.
        ------------------     --------------------------------------------------------------------
        cleanup_mode           Optional string. Defines if files in the server directory needs to
                               be cleaned up. The default is None.
        ------------------     --------------------------------------------------------------------
        description            Optional string. An optional description for the server directory.
                               The default is None.
        ==================     ====================================================================


        :return:
             A boolean indicating success (True).

        """
        url = self._url + "/directories/register"
        params = {
            "name" : name,
            "physicalPath" : physical_path,
            "directoryType" : directory_type,
            "cleanupMode" : cleanup_mode,
            "maxFileAge" : max_age
        }
        if description:
            params['description'] = description
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res

    #----------------------------------------------------------------------
    @property
    def licenses(self):
        """
        Gets the license resource list.  The licenses resource lists the
        current license level of ArcGIS for Server and all authorized
        extensions. Contact Esri Customer Service if you have questions
        about license levels or expiration properties.
        """
        url = self._url + "/licenses"
        params = {
            "f" : "json"
        }
        return self._con.get(path=url,
                             postdata=params)
    #----------------------------------------------------------------------
    @property
    def platform_services(self):
        """
        Provides access to the platform services that are associated with GeoAnalytics.
        """
        url = self._url + "/platformservices"
        return PlatformServiceManager(url=url, connection=self._con)

    #----------------------------------------------------------------------
    @property
    def jobs(self):
        """
        Gets the Jobs object.
        """
        url = self._url + "/jobs"
        return Jobs(url=url,
                    connection=self._con,
                    initialize=True)
    #----------------------------------------------------------------------
    @property
    def web_adaptors(self):
        """
        Gets a list of all the Web Adaptors that have been registered
        with the site. The server will trust all these Web Adaptors and
        will authorize calls from these servers.

        To configure a new Web Adaptor with the server, you'll need to use
        the configuration web page or the command line utility. For full
        instructions, see Configuring the Web Adaptor after installation.
        """
        url = self._url + "/webadaptors"
        params = {
            "f" : "json"
        }
        return self._con.get(path=url,
                             params=params)
    #----------------------------------------------------------------------
    @property
    def web_adaptors_configuration(self):
        """
        Gets the Web Adaptors configuration which is a resource of all the
        configuration parameters shared across all the Web Adaptors in the
        site. Most importantly, this resource lists the shared key that is
        used by all the Web Adaptors to encrypt key data bits for the
        incoming requests to the server.
        """
        url = self._url + "/webadaptors/config"
        params = {
            "f" : "json"
        }
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def update_web_adaptors_configuration(self, config):
        """
        You can use this operation to change the Web Adaptor configuration
        and the sharedkey attribute. The sharedkey attribute must be present
        in the request.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        config                 Required string. The configuration items to be updated for this web
                               adaptor. Always include the web adaptor's sharedkey attribute.
        ==================     ====================================================================

        :return:
            A boolean indicating success (True), else a Python dictionary containing an error message.

        """
        url = self._url + "/webadaptors/config/update"
        params = {
            "f" : "json",
            "webAdaptorConfig" : config
        }
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def update_web_adaptor(self,
                           wa_id,
                           description,
                           http_port,
                           https_port):
        """
        This operation allows you to update the description, HTTP port, and
        HTTPS port of a Web Adaptor that is registered with the server.

        .. note::
            This operation is only meant to change the descriptive properties
            of the Web Adaptor and does not affect the configuration of the web
            server that deploys your Web Adaptor.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        wa_id                  Required string. The web adaptor ID.
        ------------------     --------------------------------------------------------------------
        description            Required string. A description of this web adaptor.
        ------------------     --------------------------------------------------------------------
        http_port              Required integer. The HTTP port of the web server.
        ------------------     --------------------------------------------------------------------
        https_port             Required integer. The HTTPS (SSL) port of the web server that runs
                               the web adaptor.
        ==================     ====================================================================


        :return:
            A boolean indicating success (True), else a Python dictionary containing an error message.

        """
        url = self._url + "/webadaptors/{w}/update".format(w=wa_id)
        params = {
            'f' : 'json',
            'description' : description,
            'httpPort' : http_port,
            'httpsPort' : https_port
        }
        res = self._con.post(path=url,
                              postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def unregister_webadaptor(self, wa_id):
        """
        Unregistering a Web Adaptor removes the Web Adaptor from the ArcGIS
        Server's trusted list. The Web Adaptor can no longer submit requests
        to the server.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        wa_id                  Required string. The web adaptor ID.
        ==================     ====================================================================

        :return:
            A boolean indicating success (True).

        """
        url = self._url + "/webadaptors/{waid}/update".format(
            waid=wa_id)
        params = {
            "f" : "json",
        }
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    @property
    def configuration_store(self):
        """
        Gets the ConfigurationStore object for this site.
        """
        url = self._url + "/configstore"

        return ConfigurationStore(url=url,
                                  connection=self._con)
    #----------------------------------------------------------------------
    def clear_cache(self):
        """
        This operation clears the cache on all REST handlers in the system.
        While the server typically manages the REST cache for you, use this
        operation to get explicit control over the cache.

        :return:
            A boolean indicating success (True).

        """
        params = {'f': 'json'}
        url = self._url + "/handlers/rest/cache/clear"
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    @property
    def deployment(self):
        """
        Gets the load balancing value for this site.  Load balancing is an
        ArcGIS Server deployment configuration resource that can
        control the load balancing functionality between GIS server
        machines. A value of True means that load balancing is disabled.

        singleClusterMode - At 10.4, in large sites with a single cluster,
        the site is configured to prevent load balancing between GIS
        server machines. This reduces network traffic between machines
        in the site and helps reduce load on your network. Upgrades
        from earlier versions will  set this property to true if the
        site uses a single cluster. Sites with multiple clusters cannot
        use singleClusterMode.

        To prevent load balancing, the following criteria must be met:

         - All machines in the site must participate in a single
           cluster. Multiple clusters cannot exist.
         - An external load balancer or ArcGIS Web Adaptor must be
           configured to forward requests to the site. If no external
           gateway exists, requests will only be handled by the
           machine designated in the request.

        To enable load balancing, set this property to false. Updating
        this property will restart all machines in the site.
        """
        url = self._url + "/deployment"
        params = {
            "f" : "json"
        }
        return self._con.get(path=url,
                             params=params)
    #----------------------------------------------------------------------
    def _edit_services_directory(self,
                                allowed_origins,
                                arcgis_com_map,
                                arcgis_com_map_text,
                                jsapi_arcgis,
                                jsapi_arcgis_css,
                                jsapi_arcgis_css2,
                                jsapi_arcgis_sdk,
                                service_dir_enabled):
        """
        Allows you to update the Services Directory configuration.  You can do such thing as
        enable or disable the HTML view of ArcGIS REST API, or adjust the JavaScript and map viewer
        previews of services in the Services Directory so that they work with your own locally
        hosted JavaScript API and map viewer.

        ====================     ====================================================================
        **Argument**             **Description**
        --------------------     --------------------------------------------------------------------
        allowed_origins          Required string. A comma-separated list of URLs of domains allowed to
                                 make requests. An asterisk (*) can be used to denote all domains.
        --------------------     --------------------------------------------------------------------
        arcgis_com_map           Required string. URL of the map viewer application used for service
                                 previews. Defaults to the ArcGIS.com map viewer but could be used
                                 to point at your own Portal for ArcGIS map viewer.
        --------------------     --------------------------------------------------------------------
        arcgis_com_map_text      Required string. The text to use for the preview link that opens
                                 the map viewer.
        --------------------     --------------------------------------------------------------------
        jsapi_arcgis             Required string. The URL of the JavaScript API to use for service
                                 previews. Defaults to the online ArcGIS API for JavaScript, but
                                 could be pointed at your own locally-installed instance of the
                                 JavaScript API.
        --------------------     --------------------------------------------------------------------
        jsapi_arcgis_css         Required string. The CSS file associated with the ArcGIS API for
                                 JavaScript. Defaults to the online Dojo tundra.css.
        --------------------     --------------------------------------------------------------------
        jsapi_arcgis_css2        Required string. An additional CSS file associated with the ArcGIS
                                 API for JavaScript. Defaults to the online esri.css.
        --------------------     --------------------------------------------------------------------
        jsapi_arcgis_sdk         Required string. The URL of the ArcGIS API for JavaScript help.
        --------------------     --------------------------------------------------------------------
        service_dir_enabled      Required string. Flag to enable/disable the HTML view of the
                                 services directory.
        ====================     ====================================================================


        :return:
            A boolean indicating success (True).

        """
        params = {
            "f" : "json",
            "allowedOrigins": allowed_origins,
            "arcgis.com.map" : arcgis_com_map,
            "arcgis.com.map.text" : arcgis_com_map_text,
            "jsapi.arcgis" : jsapi_arcgis,
            "jsapi.arcgis.css" : jsapi_arcgis_css,
            "jsapi.arcgis.css2" : jsapi_arcgis_css2,
            "jsapi.arcgis.sdk" : jsapi_arcgis_sdk,
            "servicesDirEnabled" : service_dir_enabled
        }
        url = self._url + "/handlers/rest/servicesdirectory/edit"
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    @property
    def handlers(self):
        """
        Gets the handler of this server. A handler exposes the GIS capabilities of ArcGIS Server through a
        specific interface/API. There are two types of handlers currently
        available in the server:

          - Rest -- Exposes the REST-ful API
          - Soap -- Exposes the SOAP API

        The Rest Handler resource exposes some of the administrative
        operations on the REST handler such as clearing the cache.
        """
        url = self._url + "/handlers"
        params = {"f" : "json"}
        return self._con.get(path=url,
                             params=params)
    #----------------------------------------------------------------------
    @property
    def rest_handler(self):
        """
        Gets a list of resources accessible throught the REST API.
        """
        url = self._url + "/handlers/rest"
        params = {'f': 'json'}
        return self._con.get(path=url,
                             params=params)
########################################################################
class PlatformServiceManager(BaseServer):
    """
    The Platform Services Manager allows you to view and manage your platform services in the ArcGIS
    Server Administrator Directory. The Compute Platform and Synchronization services are used by
    GeoAnalytics Server. For each platform service, you can start it, stop it, and view its status.
    While a health check operation is exposed for all three platform services, only Compute Platform
    Health Check returns valuable information.


    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string. The service URL.
    ------------------     --------------------------------------------------------------------
    connection             Required ServerConnection. The connection object.
    ------------------     --------------------------------------------------------------------
    initialize             Optional string. Denotes whether to load the service properties at
                           creation (True). Default is False.
    ==================     ====================================================================

    """
    _cp = None
    _mb = None
    _ss = None
    _con = None
    _url = None
    _json = None
    _json_dict = None
    #----------------------------------------------------------------------
    def __init__(self,
                 url,
                 connection,
                 initialize=False):
        """
        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        url                    Required string. The service URL.
        ------------------     --------------------------------------------------------------------
        connection             Required ServerConnection. The connection object.
        ------------------     --------------------------------------------------------------------
        initialize             Optional string. Denotes whether to load the service properties at
                               creation (True). Default is False.
        ==================     ====================================================================

        """
        super(PlatformServiceManager, self).__init__(connection=connection,
                                                     url=url)
        self._url = url
        self._con = connection
        if initialize:
            self._init(connection)
    #----------------------------------------------------------------------
    def _init(self, connection=None):
        """loads the properties into the class"""
        from arcgis._impl.common._mixins import PropertyMap
        if connection is None:
            connection = self._con
        params = {"f":"json"}
        try:
            result = connection.get(path=self._url,
                                    params=params)
            if isinstance(result, dict):
                self._json_dict = result
                self._properties = PropertyMap(result)
            else:
                self._json_dict = {}
                self._properties = PropertyMap({})
        except:
            self._json_dict = {}
            self._properties = PropertyMap({})
    #----------------------------------------------------------------------
    def get(self, service):
        """
        Returns a single instance of a Platform Service

        ================   =================================================
        **Arguements**     **Description**
        ----------------   -------------------------------------------------
        service            optional string. This is the name of the service
                           that you want to examine.
                           Allowed values: SYNCHRONIZATION_SERVICE,
                           MESSAGE_BUS, or COMPUTE_PLATFORM
        ================   =================================================

        :returns PlatformService objects

        """
        services = []

        if "platformservices" in self._json_dict:
            for ps in self._json_dict["platformservices"]:
                if ps['type'].lower() == service.lower():
                    return PlatformService(url="%s/%s" % (self._url, ps['id']),
                                           gis=self._con)
        return None
    #----------------------------------------------------------------------
    def list(self):
        """
        Returns all Platform Services on the enterprise configuration.

        :returns list of PlatformService objects

        """
        services = []
        service = None
        if service is None:
            if "platformservices" in self._json_dict:
                for ps in self._json_dict["platformservices"]:
                    services.append(PlatformService(url="%s/%s" % (self._url, ps['id']),
                                                    gis=self._con))
        else:
            if "platformservices" in self._json_dict:
                for ps in self._json_dict["platformservices"]:
                    if ps['type'].lower() == service.lower():
                        services.append(PlatformService(url="%s/%s" % (self._url, ps['id']),
                                                    gis=self._con))
        return services
########################################################################
class PlatformService(BaseServer):
    """
    The platform service can be a message_bus, computre_platform, or synchronization_service.
    Each platform service has the following operations: start, stop, status, and health check.

    - The Compute Platform resource allows you to view and manage your compute platform service for ArcGIS GeoAnalytics Server in the ArcGIS Server Administrator Directory. You can start or stop the service, view its status, and run a health check.
    - The Message Bus resource allows you to view and manage your message bus service in the ArcGIS Server Administrator Directory. You can start or stop the service, and view its status. While a health check operation is exposed, it will not return any valuable information.
    - The Synchronization Service resource allows you to view and manage your synchronization service for ArcGIS GeoAnalytics Server in the ArcGIS Server Administrator Directory. You can start or stop the service and view its status. While a health check operation is exposed for the synchronization service, it will not return any valuable information.


    """
    _url = None
    _con = None
    #----------------------------------------------------------------------
    def __init__(self, url, connection, initialize=False):
        """
        Constructor

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        url                    Required string. The service URL.
        ------------------     --------------------------------------------------------------------
        connection             Required string. The connection string.
        ------------------     --------------------------------------------------------------------
        initialize             Optional string. Denotes whether to load the properties at creation
                               (True). Default is False.
        ==================     ====================================================================


        """
        initialize = True
        self._url = url
        self._con = connection
        if initialize:
            self._init(connection)
    #----------------------------------------------------------------------
    def start(self):
        """
        The Start method allows for the running of the service.
        """
        params = {'f': "json"}
        url = "%s/start" % self._url
        return self._con.get(url, params)
    #----------------------------------------------------------------------
    def stop(self):
        """
        The Start method allows for the running of the service.
        """
        params = {'f': "json"}
        url = "%s/stop" % self._url
        return self._con.get(url, params)
    #----------------------------------------------------------------------
    def status(self):
        """
        The status resource allows you to view the status of the service.
        The
        """
        params = {'f': "json"}
        url = "%s/status" % self._url
        return self._con.get(url, params)
    #----------------------------------------------------------------------
    def health(self):
        """
        The health check operation allows you to view the health of the service.
        """
        params = {'f': "json"}
        url = "%s/health" % self._url
        return self._con.get(url, params)
########################################################################
class ConfigurationStore(BaseServer):
    """
    A utility class for managing the Configuration Store of this server.
    """
    _con = None
    _url = None
    _json = None
    _json_dict = None
    #----------------------------------------------------------------------
    def __init__(self,
                 url,
                 connection,
                 initialize=False):
        """
        Constructor

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        url                    Required string. The machine URL.
        ------------------     --------------------------------------------------------------------
        connection             Required string. The connection string.
        ------------------     --------------------------------------------------------------------
        initialize             Optional string. Denotes whether to load the machine properties at
                               creation (True). Default is False.
        ==================     ====================================================================

        """
        super(ConfigurationStore, self).__init__(connection=connection,
                                                 url=url)
        self._url = url
        self._con = connection
        if initialize:
            self._init(connection)
    #----------------------------------------------------------------------
    def recover(self):
        """
        Recovers the Configuration Store of the site.

        If the shared configuration store for a site is unavailable, a site
        in read-only mode will operate in a degraded capacity that allows
        access to the ArcGIS Server Administrator Directory. You can recover
        a site if the shared configuration store is permanently lost. The
        site must be in read-only mode, and the site configuration files
        must have been copied to the local repository when switching site
        modes. The recover operation will copy the configuration store from
        the local repository into the shared configuration store location.
        The copied local repository will be from the machine in the site
        where the recover operation is performed.

        :return:
            A boolean indicating success (True).

        """
        url = self._url + "/recover"
        params = {"f" : "json"}
        return self._con.get(path=url,
                             params=params)
    #----------------------------------------------------------------------
    def edit(self,
             type_value,
             connection,
             move=True,
             run_async=False):
        """
        You can use this operation to update the configuration store.
        Typically, this operation is used to change the location of the
        configuration store.

        When ArcGIS Server is installed, the default configuration store
        uses local paths. As the site grows (more server machines are
        added), the location of the store must be updated to use a shared
        file system path. On the other hand, if you know at the onset that
        your site will have two or more server machines, you can start from
        a shared path while creating a site and skip this step altogether.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        type_value             Required string. The type of the configuration store. Values: FILESYSTEM
        ------------------     --------------------------------------------------------------------
        gis                    Required string. A file path or connection URL to the physical
                               location of the store.
        ------------------     --------------------------------------------------------------------
        move                   Optional string. A boolean to indicate if you want
                               to move the content of the current store to the new store. The
                               default True (move the content).
        ------------------     --------------------------------------------------------------------
        run_async              Optional string. Determines if this operation must run asynchronously.
                               The default is False (doesn not have to run asynchronously).
        ==================     ====================================================================


        :return:
            A boolean indicating success (True).

        """
        url = self._url + "/edit"
        params = {
            "f" : "json",
            "type" : type_value,
            "connectionString" : connection,
            "move" : move,
            "runAsync" : run_async
        }
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
########################################################################
class Jobs(BaseServer):
    """
    This resource is a collection of all the administrative jobs
    (asynchronous operations) created within your site. When operations
    that support asynchronous execution are run, the server creates a new
    job entry that can be queried for its current status and messages.
    """
    _con = None
    _json = None
    _jobs = None
    _json_dict = None
    _url = None
    #----------------------------------------------------------------------
    def __init__(self, url,
                 connection,
                 initialize=False):
        """
        Constructor

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        url                    Required string. The machine URL.
        ------------------     --------------------------------------------------------------------
        connection             Required string. The connection string.
        ------------------     --------------------------------------------------------------------
        initialize             Optional string. Denotes whether to load the machine properties at
                               creation (True). Default is False.
        ==================     ====================================================================

        """
        super(Jobs, self).__init__(connection=connection,
                                   url=url)
        self._url = url
        self._con = connection
        if initialize:
            self._init(connection)
    #----------------------------------------------------------------------
    @property
    def jobs(self):
        """
        Gets the job IDs.
        """
        if self._jobs is None:
            self._init()
        return self._jobs
    #----------------------------------------------------------------------
    def get(self, job_id):
        """
        A job represents the asynchronous execution of an operation. You
        can acquire progress information by periodically querying the job.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        job_id                 Required string. The ID of the job.
        ==================     ====================================================================

        :return:
            A JSON dictionary containing progress information for the job ID.

        """
        url = self._url + "/%s" % job_id
        params = {
            "f" : "json"
        }
        return self._con.get(path=url,
                             params=params)
########################################################################
class ServerProperties(BaseServer):
    """
    The Server has configuration parameters that can govern some of its
    intricate behavior. This Server Properties resource is a container for
    such properties. These properties are available to all server objects
    and extensions through the server environment interface.

    The properties include:

      - CacheSizeForSecureTileRequests -- An integer that specifies the
        number of users whose token information will be cached. This
        increases the speed of tile retrieval for cached services. If not
        specified, the default cache size is 200,000. Both REST and SOAP
        services honor this property. You'll need to manually restart
        ArcGIS Server in order for this change to take effect.
      - DisableAdminDirectoryCache -- Disables browser caching of the
        Administrator Directory pages. The default is False. To disable
        browser caching, set this property to True.
      - disableIPLogging -- When a possible cross-site request forgery
        (CSRF) attack is detected, the server logs a message containing
        the possible IP address of the attacker. If you do not want IP
        addresses listed in the logs, set this property to True. Also,
        HTTP request referrers are logged at FINE level by the REST and
        SOAP handlers unless this property is set to True.
      - javaExtsBeginPort -- Specifies a start port of the port range used
        for debugging Java server object extensions.
        Example: 8000
      - javaExtsEndPort -- Specifies an end port of the port range used for
        debugging Java server object extensions.
        Example: 8010
      - localTempFolder -- Defines the local folder on a machine that can
        be used by GIS services and objects. If this property is not
        explicitly set, the services and objects will revert to using the
        system's default temporary directory.

        .. note::
            If this property is used, you must create the temporary directory
            on every server machine in the site. Example: /tmp/arcgis.

      - messageFormat -- Defines the transmission protocol supported by
        the services catalog in the server.
        Values:
               - esriServiceCatalogMessageFormatBin,
               - esriServiceCatalogMessageFormatSoap,
               - esriServiceCatalogMessageFormatSoapOrBin

      - messageVersion -- Defines the version supported by the services
        catalog in the server. Example: esriArcGISVersion101
      - PushIdentityToDatabase -- Propogates the credentials of the logged-in
        user to make connections to an Oracle database. This
        property is only supported for use with Oracle databases.
        Values: True | False
      - suspendDuration -- Specifies the duration for which the ArcGIS
        service hosting processes should suspend at startup. This
        duration is specified in milliseconds. This is an optional
        property that takes effect when suspendServiceAtStartup is set
        to True. If unspecified and suspension of service at startup is
        requested, then the default suspend duration is 30 seconds.
        Example: 10000 (meaning 10 seconds)
      - suspendServiceAtStartup -- Suspends the ArcGIS service hosting
        processes at startup. This will enable attaching to those
        processes and debugging code that runs early in the lifecycle of
        server extensions soon after they are instantiated.
        Values: True | False
      - uploadFileExtensionWhitelist -- This specifies what files are
        allowed to be uploaded through the file upload API by
        identifying the allowable extensions. It is a list of comma-separated
        extensions without dots. If this property is not
        specified, a default list is used. This is the default list: soe,
        sd, sde, odc, csv, txt, zshp, kmz, and geodatabase.

        .. note::
            Updating this list overrides the default list completely. This
            means if you set this property to a subset of the default list
            then only those items in the subset will be accepted for upload.
            Example: sd, so, sde, odc.

      - uploadItemInfoFileExtensionWhitelist -- This specifies what files
        are allowed to be uploaded through the service iteminfo upload
        API by identifying the allowable extensions. It should be a list
        of comma-separated extensions without dots. If this property is
        not specified, a default list is used. This is the default list:
        xml, img, png, gif, jpg, jpeg, bmp.

        .. note::
            This list overrides the default list completely. This means if you
            set this property to a subset of the default list then only those
            items in the subset will be accepted for upload. Example: png, svg,
            gif, jpg, tiff, bmp.

      - WebContextURL -- Defines the web front end as seen by your users.
        Example: http://mycompany.com/gis

    """
    _con = None
    _url = None
    _json = None
    _json_dict = None
    #----------------------------------------------------------------------
    def __init__(self,
                 url,
                 connection,
                 initialize=False):
        """
        Constructor

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        url                    Required string. The machine URL.
        ------------------     --------------------------------------------------------------------
        connection             Required string. The connection string.
        ------------------     --------------------------------------------------------------------
        initialize             Optional string. Denotes whether to load the machine properties at
                               creation (True). Default is False.
        ==================     ====================================================================

        """
        super(ServerProperties, self).__init__(connection=connection,
                                               url=url)
        if url.lower().endswith('/properties'):
            self._url = url
        else:
            self._url = url + "/properties"
        if initialize:
            self._init(connection)
    #----------------------------------------------------------------------
    def __str__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    def update(self, properties):
        """
        This operation allows you to update the server properties. See the ServerProperties
        class description for all possible properties.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        properties             Required string. A Python dictionary of server properties to be updated.
        ==================     ====================================================================

        :return:
            A boolean indicating success (True).

        """
        url = self._url + "/update"
        params = {
            "f" : "json",
            "properties" : properties
        }
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
########################################################################
class DirectoryManager(object):
    """
    A collection of all the server directories is listed under this
    resource.

    You can add a new directory using the Register Directory operation. You
    can then configure GIS services to use one or more of these
    directories. If you no longer need the server directory, you must
    remove the directory by using the Unregister Directory operation.
    """
    _system = None
    def __init__(self, system):
        self._system = system
    #----------------------------------------------------------------------
    def __str__(self):
        return '<%s at %s>' % (type(self).__name__, self._system._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return '<%s at %s>' % (type(self).__name__, self._system._url)
    #----------------------------------------------------------------------
    def all(self):
        """
        Provides a configuration of this server directory.
        Server directories are used by GIS services as a location to output
        items such as map images, tile caches, and geoprocessing results.
        In addition, some directories contain configurations that power the
        GIS services.

        :return:
            A dictionary of this server directory configuration properties.
        """
        return self._system._directories
    #----------------------------------------------------------------------
    def edit_services_directory(self,
            allowedOrigins,
            arcgis_com_map,
            arcgis_com_map_text,
            jsapi_arcgis,
            jsapi_arcgis_css,
            jsapi_arcgis_css2,
            jsapi_arcgis_sdk,
            serviceDirEnabled):
        """
        Allows you to update the Services Directory configuration.  You can do such thing as
        enable or disable the HTML view of ArcGIS REST API, or adjust the JavaScript and map viewer
        previews of services in the Services Directory so that they work with your own locally
        hosted JavaScript API and map viewer.


        ====================     ====================================================================
        **Argument**             **Description**
        --------------------     --------------------------------------------------------------------
        allowed_origins          Required string. A comma-separated list of URLs of domains allowed to
                                 make requests. An asterisk (*) can be used to denote all domains.
        --------------------     --------------------------------------------------------------------
        arcgis_com_map           Required string. URL of the map viewer application used for service
                                 previews. Defaults to the ArcGIS.com map viewer but could be used
                                 to point at your own Portal for ArcGIS map viewer.
        --------------------     --------------------------------------------------------------------
        arcgis_com_map_text      Required string. The text to use for the preview link that opens
                                 the map viewer.
        --------------------     --------------------------------------------------------------------
        jsapi_arcgis             Required string. The URL of the JavaScript API to use for service
                                 previews. Defaults to the online ArcGIS API for JavaScript, but
                                 could be pointed at your own locally-installed instance of the
                                 JavaScript API.
        --------------------     --------------------------------------------------------------------
        jsapi_arcgis_css         Required string. The CSS file associated with the ArcGIS API for
                                 JavaScript. Defaults to the online Dojo tundra.css.
        --------------------     --------------------------------------------------------------------
        jsapi_arcgis_css2        Required string. An additional CSS file associated with the ArcGIS
                                 API for JavaScript. Defaults to the online esri.css.
        --------------------     --------------------------------------------------------------------
        jsapi_arcgis_sdk         Required string. The URL of the ArcGIS API for JavaScript help.
        --------------------     --------------------------------------------------------------------
        service_dir_enabled      Required string. Flag to enable/disable the HTML view of the
                                 services directory.
        ====================     ====================================================================


        :return:
            A boolean indicating success (True).

        """
        return self._system._edit_services_directory(allowedOrigins, arcgis_com_map,
                                                     arcgis_com_map_text,
                                                     jsapi_arcgis,
                                                     jsapi_arcgis_css,
                                                     jsapi_arcgis_css2,
                                                     jsapi_arcgis_sdk,
                                                     serviceDirEnabled)
    #----------------------------------------------------------------------
    def get(self, name):
        """
        Retrieves a single directory registered with ArcGIS Server.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        name                   Required string. The name of the registered directory.
        ==================     ====================================================================

        :return:
            The directory object.
        """
        return self._system._get_directory(name=name)
    #----------------------------------------------------------------------
    def add(self,
            name,
            physicalPath,
            directoryType,
            maxFileAge,
            cleanupMode="NONE",
            description=None):
        """
        Registers a new server directory. While registering the server
        directory, you can also specify the directory's cleanup parameters.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        name                   Required string. The name of the server directory.
        ------------------     --------------------------------------------------------------------
        physical_path          Required string. The absolute physical path of the server directory.
        ------------------     --------------------------------------------------------------------
        directory_type         Required string. The type of server directory.
        ------------------     --------------------------------------------------------------------
        max_age                Required integer. The length of time a file in the directory needs
                               to be kept before it is deleted.
        ------------------     --------------------------------------------------------------------
        cleanup_mode           Optional string. Defines if files in the server directory needs to
                               be cleaned up. The default is None.
        ------------------     --------------------------------------------------------------------
        description            Optional string. An optional description for the server directory.
                               The default is None.
        ==================     ====================================================================

        :return:
            A boolean indicating success (True).

        """
        return self._system._register(name, physicalPath, directoryType,
                                      maxFileAge, cleanupMode, description)
########################################################################
class ServerDirectory(BaseServer):
    """
    Server directories are used by GIS services as a location to output
    items such as map images, tile caches, and geoprocessing results. In
    addition, some directories contain configurations that power the GIS
    services.
    In a Site with more than one server machine these directories must be
    available on network shares, accessible to every machine in the site.

    The following directory types can be registered with the server:

     - Output -- Stores various information generated by services, such as map
       images. Instances: One or more
     - Cache -- Stores tile caches used by map, globe, and image services for
       rapid performance. Instances: One or more
     - Jobs -- Stores results and other information from geoprocessing
       services. Instances: One or more
     - System -- Stores files that are used internally by the GIS server.
       Instances: One

    Server directories that contain output of various GIS
    services can be periodically cleaned to remove old unused files. By
    using the cleanup mode and maximum file age parameters, you control
    when when you would like the files in these directories to be
    cleaned.

    All the output server directories are automatically virtualized (they
    can be accessed over a URL) for you through the ArcGIS Server REST API.
    """
    _con = None
    _url = None
    _json = None
    _json_dict = None
    _url = None
    _name = None
    _physicalPath = None
    _directoryType = None
    _cleanupMode = None
    _maxFileAge = None
    _description = None
    _virtualPath = None
    #----------------------------------------------------------------------
    def __init__(self,
                 url,
                 connection,
                 initialize=False):
        """
        Constructor

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        url                    Required string. The machine URL.
        ------------------     --------------------------------------------------------------------
        connection             Required string. The connection string.
        ------------------     --------------------------------------------------------------------
        initialize             Optional string. Denotes whether to load the machine properties at
                               creation (True). Default is False.
        ==================     ====================================================================

        """
        super(ServerDirectory, self).__init__(connection=connection,
                                              url=url)
        self._url = url
        self._con = connection
        if initialize:
            self._init(connection)
    #----------------------------------------------------------------------
    def edit(self,
             physical_path,
             cleanup_mode,
             max_age,
             description):
        """
        The server directory's edit operation allows you to change the path
        and clean up properties of the directory. This operation updates
        the GIS service configurations (and points them to the new path)
        that are using this directory, causing them to restart. It is
        therefore recommended that any edit to the server directories be
        performed when the server is not under load.

        This operation is mostly used when growing a single machine site to
        a multiple machine site configuration, which requires that the
        server directories and configuration store be put on a
        network-accessible file share.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        physical_path          Required string. The absolute physical path of the server directory.
        ------------------     --------------------------------------------------------------------
        cleanup_mode           Optional string. Defines if files in the server directory needs to
                               be cleaned up. The default is None.
        ------------------     --------------------------------------------------------------------
        max_age                Required integer. The length of time a file in the directory needs
                               to be kept before it is deleted.
        ------------------     --------------------------------------------------------------------
        description            Optional string. An optional description for the server directory.
                               The default is None.
        ==================     ====================================================================


        :return:
            A boolean indicating success (True).

        """
        url = self._url + "/edit"
        params = {
            "f" : "json",
            "physicalPath" : physical_path,
            "cleanupMode" : cleanup_mode,
            "maxFileAge" : max_age,
            "description" : description
        }
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def clean(self):
        """
        Cleans the content (files and folders) within the directory that
        have passed their expiration date. Every server directory has the
        max file age and cleanup mode parameter that govern when a file
        created inside is supposed to be cleaned up. The server directory
        cleaner automatically cleans up the content within server
        directories at regular intervals. However, you can explicitly clean
        the directory by invoking this operation.

        :return:
            A boolean indicating success (True).

        """
        url = self._url + "/clean"
        params = {
            "f" : "json"
        }
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def recover(self):
        """
        Recovers the shared server directories of the site.

        If the shared server directories for a site are unavailable, a site
        in read-only mode will operate in a degraded capacity that allows
        access to the ArcGIS Server Administrator Directory. You can recover
        a site if the shared server directories are permanently lost. The
        site must be in read-only mode, and the site configuration files
        must have been copied to the local repository when switching site
        modes. The recover operation will copy the server directories from
        the local repository into the shared server directories location.
        The copied local repository will be from the machine in the site
        where the recover operation is performed.

        :return:
            A boolean indicating success (True).

        """
        url = self._url + "/recover"
        params = {'f': 'json'}
        return self._con.get(path=url,
                             params=params)
    #----------------------------------------------------------------------
    def unregister(self):
        """
        Unregisters a server directory. Once a directory has been
        unregistered, it can no longer be referenced (used) from within a
        GIS service.

        :return:
            A boolean indicating success (True).

        """
        url = self._url + "/unregister"
        params = {
            "f" : "json"
        }
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res



"""
In the GIS, entities located in space with a set of properties can be represented as features. This module has the types
to represent features and collection of features.
"""
import copy
import json
import os
import re
import tempfile
import uuid
from datetime import datetime
from arcgis._impl.common._mixins import PropertyMap
from arcgis._impl.common._spatial import json_to_featureclass
from arcgis._impl.common._utils import _date_handler
from arcgis.geometry import BaseGeometry, Point, MultiPoint, Polyline, Polygon, Geometry, SpatialReference
from arcgis.gis import Layer

try:
    import arcpy
    HASARCPY = True
except ImportError:
    HASARCPY = False

class Feature(object):
    """ Entities located in space with a set of properties can be represented as features. """
    _geom = None
    _json = None
    _dict = None
    _geom_type = None
    _attributes = None
    _wkid = None

    # ----------------------------------------------------------------------
    def __init__(self, geometry=None, attributes=None):
        """Constructor"""
        self._dict = {}
        if geometry is not None:
            self._dict["geometry"] = geometry
        if attributes is not None:
            self._dict["attributes"] = attributes

    # ----------------------------------------------------------------------
    def set_value(self, field_name, value):
        """
        Sets an attribute value for a given field name

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        field_name          Required String. The name of the field to update.
        ---------------     --------------------------------------------------------------------
        value               Required object. The value to update the field with.
        ===============     ====================================================================


        :return: boolean

        """
        if field_name in self.fields:
            if value is not None:
                self._dict['attributes'][field_name] = value
                self._json = json.dumps(self._dict, default=_date_handler)
            else:
                pass
        elif field_name.upper() in ['SHAPE', 'SHAPE@', "GEOMETRY"]:
            if isinstance(value, BaseGeometry):
                if isinstance(value, Point):
                    self._dict['geometry'] = {
                        "x": value['x'],
                        "y": value['y']
                    }
                elif isinstance(value, MultiPoint):
                    self._dict['geometry'] = {
                        "points": value['points']
                    }
                elif isinstance(value, Polyline):
                    self._dict['geometry'] = {
                        "paths": value['paths']
                    }
                elif isinstance(value, Polygon):
                    self._dict['geometry'] = {
                        "rings": value['rings']
                    }
                else:
                    return False
                self._json = json.dumps(self._dict, default=_date_handler)
        else:
            return False
        return True

    # ----------------------------------------------------------------------
    def get_value(self, field_name):
        """
        returns a value for a given field name

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        field_name          Required String. The name of the field to get the value for.
        ===============     ====================================================================


        :return: object

        """
        if field_name in self.fields:
            return self._dict['attributes'][field_name]
        elif field_name is not None and field_name.upper() in ['SHAPE', 'SHAPE@', "GEOMETRY"]:
            return self._dict['geometry']
        return None

    # ----------------------------------------------------------------------
    @property
    def as_dict(self):
        """returns the feature as a dictionary"""
        return self._dict

    # ----------------------------------------------------------------------
    @property
    def as_row(self):
        """ converts a feature to a list for insertion into an insert cursor
            :return:
               [row items], [field names]
               returns a list of fields and the row object
        """
        fields = self.fields
        row = [""] * len(fields)
        for key, val in self._attributes.items():
            row[fields.index(key)] = val
            del val
            del key
        if self.geometry is not None:
            row.append(self.geometry)
            fields.append("SHAPE@")
        return row, fields

    # ----------------------------------------------------------------------
    @property
    def geometry(self):
        """returns the feature geometry"""
        if self._geom is None:
            if 'geometry' in self._dict.keys():
                self._geom = self._dict['geometry']
            else:
                return None
        return self._geom

    @geometry.setter
    def geometry(self, value):
        """gets/sets a feature's geometry"""
        self._geom = value
        self._dict['geometry'] = value

    # ----------------------------------------------------------------------
    @property
    def attributes(self):
        """returns the feature attributes"""
        if self._attributes is None and 'attributes' in self._dict:
            self._attributes = self._dict['attributes']
        return self._attributes

    @attributes.setter
    def attributes(self, value):
        """gets/sets a feature's attributes"""
        self._attributes = value
        self._dict['attributes'] = value

    # ----------------------------------------------------------------------
    @property
    def fields(self):
        """ returns a list of feature fields """
        if 'attributes' in self._dict:
            self._attributes = self._dict['attributes']
            return list(self._attributes.keys())
        else:
            return []

    # ----------------------------------------------------------------------
    @property
    def geometry_type(self):
        """ returns the feature's geometry type """
        if self._geom_type is None:
            if self.geometry is not None:
                self._geom_type = self.geometry.type
            else:
                self._geom_type = "Table"
        return self._geom_type

    # ----------------------------------------------------------------------
    @classmethod
    def from_json(cls, json_str):
        """returns a feature from a JSON string"""
        feature = json.loads(json_str)
        geom = feature['geometry'] if 'geometry' in feature else None
        attribs = feature['attributes'] if 'attributes' in feature else None
        return cls(geom, attribs)

    # ----------------------------------------------------------------------
    @classmethod
    def from_dict(cls, feature):
        """returns a feature from a dict"""
        geom = feature['geometry'] if 'geometry' in feature else None
        attribs = feature['attributes'] if 'attributes' in feature else None
        if 'centroid' in feature:
            if attribs is None:
                attribs = {'centroid' : feature['centroid']}
            elif 'centroid' in attribs:
                fld = "centroid_" + uuid.uuid4().hex[:2]
                attribs[fld] = feature['centroid']
            else:
                attribs['centroid'] = feature['centroid']
        return cls(geom, attribs)
    # ----------------------------------------------------------------------
    def __str__(self):
        """"""
        return json.dumps(self.as_dict, default=_date_handler)

    __repr__ = __str__


class FeatureSet(object):
    """
    A set of features with information about their fields, field aliases, geometry type, spatial reference etc.

    FeatureSets are commonly used as input/output with several Geoprocessing
    Tools, and can be the obtained through the query() methods of feature layers.
    A FeatureSet can be combined with a layer definition to compose a FeatureCollection.

    FeatureSet contains Feature objects, including the values for the
    fields requested by the user. For layers, if you request geometry
    information, the geometry of each feature is also returned in the
    FeatureSet. For tables, the FeatureSet does not include geometries.

    If a Spatial Reference is not specified at the FeatureSet level, the
    FeatureSet will assume the SpatialReference of its first feature. If
    the SpatialReference of the first feature is also not specified, the
    spatial reference will be UnknownCoordinateSystem.
    """
    _fields = None
    _features = None
    _has_z = None
    _has_m = None
    _geometry_type = None
    _spatial_reference = None
    _object_id_field_name = None
    _global_id_field_name = None
    _display_field_name = None
    _allowed_geom_types = ["esriGeometryPoint", "esriGeometryMultipoint", "esriGeometryPolyline",
                           "esriGeometryPolygon", "esriGeometryEnvelope"]

    # ----------------------------------------------------------------------
    def __init__(self,
                 features,
                 fields=None,
                 has_z=False,
                 has_m=False,
                 geometry_type=None,
                 spatial_reference=None,
                 display_field_name=None,
                 object_id_field_name=None,
                 global_id_field_name=None):
        """Constructor"""
        self._fields = fields

        self._has_z = has_z
        self._has_m = has_m
        self._geometry_type = geometry_type
        self._spatial_reference = spatial_reference
        self._display_field_name = display_field_name
        self._object_id_field_name = object_id_field_name
        self._global_id_field_name = global_id_field_name

        # conversion of different inputs to a common list of feature objects
        if isinstance(features, str):
            # convert the featuresclass to a list of features
            features = self._fc_to_features(dataset=features)
            if features is None:
                raise AttributeError("Feature class could not be converted to a feature set")
        elif isinstance(features, list) and len(features) > 0:
            feature = features[0]
            if isinstance(feature, Feature):
                pass
                # features passed in as a list of Feature objects

                # if "attributes" in feature.as_dict:
                #     if "geometry" in feature.as_dict:
                #         features = [Feature(feat.as_dict['geometry'], feat.as_dict['attributes']) for feat in features]
                #     else:
                #         features = [Feature(None, feat.as_dict['attributes']) for feat in features]
                # elif "geometry" in feature.as_dict:
                #     features = [Feature(feat.as_dict['geometry'], None) for feat in features]
            elif isinstance(feature, dict):
                # features passed in as a list of dicts
                if "attributes" in feature:
                    if "geometry" in feature:
                        features = [Feature(feat['geometry'], feat['attributes']) for feat in features]
                    else:
                        features = [Feature(None, feat['attributes']) for feat in features]
                elif "geometry" in feature:
                    features = [Feature(feat['geometry'], None) for feat in features]
            else:
                raise AttributeError("FeatureSet requires a list of features (as dicts or Feature objects)")

        self._features = features
        if len(features) > 0:
            feat_geom = None
            feature = features[0]

            if "geometry" in feature.as_dict: # can construct features out of tables with just attributes, no geometry
                feat_geom = feature.geometry
            elif isinstance(feature, dict):
                if "geometry" in feature:
                    feat_geom = feature['geometry']

            if feat_geom is not None:
                if spatial_reference is None:
                    if 'spatialReference' in feat_geom:
                        self._spatial_reference = feat_geom['spatialReference']

                if isinstance(feat_geom, Geometry):
                    geometry = feat_geom
                else:
                    geometry = Geometry(feat_geom)

                if geometry_type is None:
                    if isinstance(geometry, Polyline):
                        self._geometry_type = "esriGeometryPolyline"
                    elif isinstance(geometry, Polygon):
                        self._geometry_type = "esriGeometryPolygon"
                    elif isinstance(geometry, Point):
                        self._geometry_type = "esriGeometryPoint"
                    elif isinstance(geometry, MultiPoint):
                        self._geometry_type = "esriGeometryMultipoint"
                # else:
                #     raise AttributeError("Invalid geometry type") # Dont raise this error as input can be tables without geometries

            # region - build fields into a dict
            if self._fields is None or len(self._fields) == 0:
                self._fields = feature.fields  # get fields from first feature if not set

            if self._fields and isinstance(self._fields[0], str):  # as in geocoded results
                _fields = []

                for key, val in feature.attributes.items():
                    if isinstance(val, float):
                        field_type = 'esriFieldTypeDouble'
                    elif isinstance(val, int):
                        field_type = 'esriFieldTypeInteger'
                    else:
                        field_type = 'esriFieldTypeString'

                    _fields.append({'name': key,
                                    'alias': key,
                                    'type': field_type,
                                    'sqlType': 'sqlTypeOther'})

                self._fields = _fields
            # endregion

            # Try to find the object ID field if not specified
            if self._object_id_field_name is None:
                # check to see if features a dict or feature object
                if isinstance(feature, Feature):
                    # Look for OBJECTID first, if it does not exist, look for FID
                    if self._fields is None or len(self._fields) == 0:
                        self._fields = feature.fields  # get fields from first feature if not set
                    for field in feature.fields:
                        if re.search("^{0}$".format("OBJECTID"), field, re.IGNORECASE):
                            self._object_id_field_name = field
                            break
                    for field in feature.fields:
                        if re.search("^{0}$".format("FID"), field, re.IGNORECASE):
                            self._object_id_field_name = field
                            break
                else:
                    for field, _ in feature.items():
                        if re.search("^{0}$".format("OBJECTID"), field, re.IGNORECASE):
                            self._object_id_field_name = field
                            break
                    for field, _ in feature.items():
                        if re.search("^{0}$".format("FID"), field, re.IGNORECASE):
                            self._object_id_field_name = field
                            break

            # if still none, then build the objectid field
            if not self._object_id_field_name:
                obj_field = {'name': 'OBJECTID',
                             'type': 'esriFieldTypeOID',
                             'alias': 'OBJECTID',
                             'sqlType': 'sqlTypeOther'
                             }

                if len(self._fields) > 0:
                    field0 = self._fields[0]
                    if isinstance(field0, str):
                        self._fields.append('OBJECTID')
                    else:
                        self._fields.append(obj_field)
                else:
                    self._fields.append(obj_field)

                self._object_id_field_name = obj_field['name']

                counter = 0
                for feat in self._features:
                    counter += 1
                    feat.fields.append('OBJECTID')
                    if not feat.attributes: #when features have just geometry and no attributes
                        feat.attributes = {'OBJECTID':counter}
                    else:
                        feat.attributes['OBJECTID'] = counter

            # rectify Object ID field type
            if self._object_id_field_name:
                for f in self._fields:
                    if f['name'] == self._object_id_field_name:
                        if f['type'] != 'esriFieldTypeOID':
                            f['type'] = 'esriFieldTypeOID'
                        break

    # ----------------------------------------------------------------------
    def __str__(self):
        """returns object as string"""
        return json.dumps(self.value, default=_date_handler)

    def __repr__(self):
        return '<{}> {} features'.format(self.__class__.__name__, len(self.features))

    # noinspection PyUnresolvedReferences
    @staticmethod
    def _fc_to_features(dataset):
        """
        Converts a dataset to a list of feature objects, if ArcPy is available

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        dataset             Required string. Path to the featureclass.
        ===============     ====================================================================


        :return: list of feature objects
        """
        try:
            import arcpy
            arcpy_found = True
        except:
            arcpy_found = False
            raise AttributeError("ArcPy is required to create a feature set from a feature class")
        if arcpy_found:
            if not arcpy.Exists(dataset=dataset):
                raise AttributeError("Error creating FeatureSet: {0} does not exist".format(dataset))

            desc = arcpy.Describe(dataset)
            fields = [field.name for field in arcpy.ListFields(dataset) if field.type not in ['Geometry']]
            date_fields = [field.name for field in arcpy.ListFields(dataset) if field.type == 'Date']
            non_geom_fields = copy.deepcopy(fields)
            features = []
            if hasattr(desc, "shapeFieldName"):
                fields.append("SHAPE@JSON")
            del desc
            with arcpy.da.SearchCursor(dataset, fields) as rows:
                for row in rows:
                    row = list(row)
                    for date_field in date_fields:
                        if row[fields.index(date_field)] is not None:
                            row[fields.index(date_field)] = int((_date_handler(row[fields.index(date_field)])))
                    template = {
                        "attributes": dict(zip(non_geom_fields, row))
                    }
                    if "SHAPE@JSON" in fields:
                        template['geometry'] = \
                            json.loads(row[fields.index("SHAPE@JSON")])

                    features.append(
                        Feature.from_dict(template)
                    )
                    del row
            return features
        return None
        # ----------------------------------------------------------------------

    @property
    def value(self):
        """returns object as dictionary"""
        val = {
            "features": [f.as_dict for f in self._features]
        }

        if self._object_id_field_name is not None:
            val["objectIdFieldName"] = self._object_id_field_name
        if self._display_field_name is not None:
            val["displayFieldName"] = self._display_field_name
        if self._global_id_field_name is not None:
            val["globalIdFieldName"] = self._global_id_field_name
        if self._spatial_reference is not None:
            val["spatialReference"] = self._spatial_reference
        if self._geometry_type is not None:
            val["geometryType"] = self._geometry_type
        if self._has_z:
            val["hasZ"] = self._has_z
        if self._has_m:
            val["hasM"] = self._has_m
        if self._fields is not None:
            val["fields"] = self._fields

        return val

        # return {
        #    "objectIdFieldName" : self._objectIdFieldName,
        #    "displayFieldName" : self._displayFieldName,
        #    "globalIdFieldName" : self._globalIdFieldName,
        #    "geometryType" : self._geometryType,
        #    "spatialReference" : self._spatialReference,
        #    "hasZ" : self._hasZ,
        #    "hasM" : self._hasM,
        #    "fields" : self._fields,
        #    "features" : [f.as_dict for f in self._features]
        # }

    # ----------------------------------------------------------------------
    @property
    def to_json(self):
        """converts the object to JSON"""
        return json.dumps(self.value, default=_date_handler)

    # ----------------------------------------------------------------------
    @property
    def to_geojson(self):
        """converts the object to GeoJSON"""
        def esri_to_geo(esrijson):
            """converts Esri Format JSON to GeoJSON"""
            def extract(feature, esri_geom_type):
                """creates a single feature"""
                item = {}
                item["type"] = "Feature"
                geom = feature["geometry"]
                geometry = {}
                geometry["type"] = get_geom_type(esri_geom_type)
                geometry["coordinates"] = get_coordinates(geom, geometry["type"])
                item["geometry"] = geometry
                item["properties"] = feature["attributes"]

                return item

            def get_geom_type(esri_type):
                """converts esri geometry types to
                GeoJSON geometry types"""
                if esri_type == "esriGeometryPoint":
                    return "Point"
                elif esri_type == "esriGeometryMultiPoint":
                    return "MultiPoint"
                elif esri_type == "esriGeometryPolyline":
                    return "LineString"
                elif esri_type == "esriGeometryPolygon":
                    return "Polygon"
                else:
                    return "Point"
            def get_coordinates(geom, geom_type):
                """
                converts the Esri Geometry Structure to
                GeoJSON structure"""
                if geom_type == "Polygon":
                    return geom["rings"]
                elif geom_type == "LineString":
                    return geom["paths"]
                elif geom_type == "Point":
                    return [ geom["x"], geom["y"] ]
                else:
                    return []
            geojson = {}
            features = esrijson["features"]
            esri_geom_type = esrijson["geometryType"]
            count = len(features)
            geojson["type"] = "FeatureCollection"
            feats = []
            for feat in features:
                feats.append(extract(feat, esri_geom_type))
            geojson["features"] = feats
            return geojson
        return json.dumps(esri_to_geo(self.value), default=_date_handler)

    def to_dict(self):
        """converts the object to Python dictionary"""
        return self.value

    @property
    def df(self):
        """converts the FeatureSet to a Pandas dataframe. Requires pandas"""
        try:
            try:
                import arcpy
                arcpy_found = True
            except:
                arcpy_found = False
            from pandas.io.json import json_normalize
            from arcgis.features import SpatialDataFrame
            if len(self.features) == 0:
                import pandas as pd
                return pd.DataFrame()
            elif self.geometry_type is not None:
                if self._spatial_reference and \
                   'wkt' in self._spatial_reference.keys():
                    sr = SpatialReference(self._spatial_reference)
                elif self._spatial_reference and \
                     'wkid' in self._spatial_reference:
                    sr = SpatialReference(self._spatial_reference)
                else:
                    sr = None
                geoms = []
                attributes = []
                for feat in self.features:
                    attributes.append(feat.attributes)
                    if isinstance(feat.geometry, Geometry):
                        geoms.append(feat.geometry)
                    else:
                        g = Geometry(feat.geometry)
                        if 'spatialReference' not in g and \
                           sr is not None:
                            g['spatialReference'] = sr
                        geoms.append(g)
                    del feat
                df = json_normalize(attributes)
                df.columns = df.columns.str.replace('attributes.', '')
                return SpatialDataFrame(df, geometry=geoms, sr=sr)
            else:
                #df = pandas.DataFrame.from_dict([f.attributes for f in fs.features])
                df = json_normalize(self.value['features'])
                df.columns = df.columns.str.replace('attributes.', '')
                if self._object_id_field_name is not None:
                    df.set_index([self._object_id_field_name], inplace=True)
                else:
                    if 'OBJECTID' in df.columns:
                        df.set_index(['OBJECTID'], inplace=True)
                    elif 'FID' in df.columns:
                        df.set_index(['FID'], inplace=True)
                return df
        except ImportError:
            raise ImportError("pandas not found, please install it")

    # ----------------------------------------------------------------------
    def __iter__(self):
        """featureset iterator on features in feature set"""
        for feature in self._features:
            yield feature

    # ----------------------------------------------------------------------
    def __len__(self):
        """returns the number of features in feature set"""
        return len(self._features)

    # ----------------------------------------------------------------------
    @staticmethod
    def from_json(json_str):
        """returns a featureset from a JSON string"""
        return FeatureSet.from_dict(json.loads(json_str))

    @staticmethod
    def from_dataframe(df):
        """returns a featureset from a Pandas' Data or Spatial DataFrame"""
        def _infer_type(df, col):
            """
            internal function used to get the datatypes for the feature class if
            the dataframe's _field_reference is NULL or there is a column that does
            not have a dtype assigned to it.

            Input:
             dataframe - spatialdataframe object
            Ouput:
              field type name
            """
            import six
            import numpy as np
            nn = df[col].notnull()
            nn = list(df[nn].index)
            if len(nn) > 0:
                val = df[col][nn[0]]
                if isinstance(val, six.string_types):
                    return "esriFieldTypeString"
                elif isinstance(val, tuple(list(six.integer_types) + [np.int32])):
                    return "esriFieldTypeInteger"
                elif isinstance(val, (float, np.int64)):
                    return "esriFieldTypeDouble"
                elif isinstance(val, datetime):
                    return "esriFieldTypeDate"
            return "esriFieldTypeString"
        from ._data.geodataset import SpatialDataFrame
        import pandas as pd
        try:
            import arcpy
            HASARCPY = True
        except ImportError:
            HASARCPY = False
        features = []
        index = 0
        sr = None
        df = df.fillna('')
        if isinstance(df, SpatialDataFrame):
            df_rows = df.copy()
            del df_rows['SHAPE']
            geoms = df['SHAPE'].tolist()
            sr = df.sr
        elif isinstance(df, pd.DataFrame):
            geoms = []
            df_rows = df.copy().to_dict('records')
        else:
            raise ValueError("Invalid input type")
        index = 0
        for row in df_rows.to_dict('records'):
            if len(geoms) > 0:
                features.append(
                    {
                        "geometry": json.loads(json.dumps(geoms[index])),
                        "attributes": row
                    })
            else:
                features.append(
                    {
                        "attributes": row
                    })
            index += 1
        fs =  FeatureSet.from_dict(featureset_dict={'features': features})
        fields = []
        for col in df_rows.columns:
            #if col not in df_rows.geometry.name:
            fields.append(
                {
                    "name" : col,
                    "type" : _infer_type(df=df_rows, col=col)
                }
            )
        fs._fields = fields
        if sr is not None:
            fs.spatial_reference = sr

        return fs
    # ----------------------------------------------------------------------
    @staticmethod
    def from_geojson(geojson):
        """
        Converts a GeoJSON Feature Collection into a FeatureSet

        """
        from warnings import warn
        def geo_to_esri(geojson):
            esri = {}

            # we already know the spatial reference we want
            # for geojson, at least in this simple case
            if 'crs' in geojson:
                warn("crs has been deprecated and will be ignored. Please"+ \
                     " see: https://tools.ietf.org/html/rfc7946#section-4 for" +\
                     " more information.")
            sr = {"wkid": 4326}

            # This will hold the geojson geometry type
            geo_type = ""
            # check for collection of features
            # and iterate as necessary
            attribute_fields = []
            if geojson["type"] == "FeatureCollection":
                features = geojson["features"]
                geo_type = features[0]["geometry"]["type"]
                attribute_fields = features[0]["properties"]

                esri_features = map(extract, features)
            else:
                attribute_fields = geojson["properties"]
                geo_type = geojson["geometry"]["type"]
                esri_features = extract(geojson)

            fields = map(extract_field, attribute_fields)
            # everything should be ready to define for the
            # esri json
            esri["geometryType"] = get_geom_type(geo_type)
            esri["spatialReference"] = sr
            esri["fields"] = list(fields)
            esri["features"] = list(esri_features)

            return esri

        def extract(feature):
            # parse out the geometry data
            geometry = get_geometry(feature)
            out_feature = {}
            out_feature["geometry"] = geometry
            out_feature["attributes"] = feature["properties"]

            return out_feature

        def extract_field(attribute):
            # now we need the fields in the properties
            a = {}
            a["alias"] = attribute
            a["name"] = attribute
            if isinstance(attribute, int):
                a["type"] = "esriFieldTypeSmallInteger"
            elif isinstance(attribute, float):
                a["type"] = "esriFieldTypeDouble"
            else:
                a["type"] = "esriFieldTypeString"
                a["length"] = 70
            return a

        def get_geom_type(geo_type):
            if geo_type == "Point":
                return "esriGeometryPoint"
            elif geo_type == "MultiPoint":
                return "esriGeometryMultiPoint"
            elif geo_type == "LineString":
                return "esriGeometryPolyline"
            elif geo_type == "Polygon" or geo_type == "MultiPolygon":
                return "esriGeometryPolygon"
            else:
                return "unknown"

        def get_geometry(feature):
            # match how geometry is represented
            # based on the geojson geometry type
            geometry = {}
            geom = feature["geometry"]
            geo_type = geom["type"]
            if geo_type == "Point":
                geometry["x"] = geom["coordinates"][0]
                geometry["y"] = geom["coordinates"][1]
            elif geo_type == "Polygon":
                geometry["rings"] = geom["coordinates"]
            elif geo_type == "MultiPolygon":
                rings = []
                if HASARCPY:
                    geom = arcpy.AsShape(geom)
                    geometry = Geometry(json.loads(geom))
                else:
                    coordkey = ([d for d in geom if d.lower() == 'coordinates']
                                    or ['coordinates']).pop()
                    coordinates = geom[coordkey]
                    typekey = ([d for d in geom if d.lower() == 'type']
                                   or ['type']).pop()
                    if geom[typekey].lower() == "polygon":
                        coordinates = [coordinates]
                    part_list = []
                    for part in coordinates:
                        part_item = []
                        for idx, ring in enumerate(part):
                            if idx:
                                part_item.append(None)
                            for coord in ring:
                                part_item.append(coord)
                        if part_item:
                            part_list.append([part_item])
                    geometry["rings"] = part_list[0]
            elif geo_type =="LineString":
                geometry["paths"] = geom

            return geometry
        return FeatureSet.from_dict(geo_to_esri(geojson))
    # ----------------------------------------------------------------------
    @staticmethod
    def from_dict(featureset_dict):
        """returns a featureset from a dict"""

        features = []
        if 'fields' in featureset_dict:
            fields = featureset_dict['fields']
        else:
            fields = []
        if 'features' in featureset_dict:
            for feat in featureset_dict['features']:
                features.append(Feature.from_dict(feat))
        return FeatureSet(
            features=features, fields=fields,
            has_z=featureset_dict['hasZ'] if 'hasZ' in featureset_dict else False,
            has_m=featureset_dict['hasM'] if 'hasM' in featureset_dict else False,
            geometry_type=featureset_dict['geometryType'] if 'geometryType' in featureset_dict else None,
            object_id_field_name=featureset_dict['objectIdFieldName'] if 'objectIdFieldName' in featureset_dict else None,
            global_id_field_name=featureset_dict['globalIdFieldName'] if 'globalIdFieldName' in featureset_dict else None,
            display_field_name=featureset_dict['displayFieldName'] if 'displayFieldName' in featureset_dict else None,
            spatial_reference=featureset_dict['spatialReference'] if 'spatialReference' in featureset_dict else None)

    # ----------------------------------------------------------------------
    @property
    def spatial_reference(self):
        """gets the featureset's spatial reference"""
        return self._spatial_reference

    # ----------------------------------------------------------------------
    @spatial_reference.setter
    def spatial_reference(self, value):
        """sets the featureset's spatial reference"""
        if isinstance(value, SpatialReference):
            self._spatial_reference = value
        elif isinstance(value, int):
            self._spatial_reference = SpatialReference(wkid=value)
        elif isinstance(value, str) and \
                str(value).isdigit():
            self._spatial_reference = SpatialReference(wkid=int(value))
        else:
            self._spatial_reference = SpatialReference(value)


    # ----------------------------------------------------------------------
    @property
    def has_z(self):
        """gets/sets the Z-property"""
        return self._has_z

    # ----------------------------------------------------------------------
    @has_z.setter
    def has_z(self, value):
        """gets/sets the Z-property"""
        if isinstance(value, bool):
            self._has_z = value

    # ----------------------------------------------------------------------
    @property
    def has_m(self):
        """gets/set the M-property"""
        return self._has_m

    # ----------------------------------------------------------------------
    @has_m.setter
    def has_m(self, value):
        """gets/set the M-property"""
        if isinstance(value, bool):
            self._has_m = value

    # ----------------------------------------------------------------------
    @property
    def geometry_type(self):
        """gets/sets the geometry Type"""
        return self._geometry_type

    # ----------------------------------------------------------------------
    @geometry_type.setter
    def geometry_type(self, value):
        """gets/sets the geometry Type"""
        if value in self._allowed_geom_types:
            self._geometry_type = value

    # ----------------------------------------------------------------------
    @property
    def object_id_field_name(self):
        """gets/sets the object id field"""
        return self._object_id_field_name

    # ----------------------------------------------------------------------
    @object_id_field_name.setter
    def object_id_field_name(self, value):
        """gets/sets the object id field"""
        self._object_id_field_name = value

    # ----------------------------------------------------------------------
    @property
    def global_id_field_name(self):
        """gets/sets the globalIdFieldName"""
        return self._global_id_field_name

    # ----------------------------------------------------------------------
    @global_id_field_name.setter
    def global_id_field_name(self, value):
        """gets/sets the globalIdFieldName"""
        self._global_id_field_name = value

    # ----------------------------------------------------------------------
    @property
    def display_field_name(self):
        """gets/sets the displayFieldName"""
        return self._display_field_name

    # ----------------------------------------------------------------------
    @display_field_name.setter
    def display_field_name(self, value):
        """gets/sets the displayFieldName"""
        self._display_field_name = value

    # ----------------------------------------------------------------------
    def save(self, save_location, out_name, encoding=None):
        """
        Saves a featureset object to a feature class




        =================    ====================================================================
        **Argument**         **Description**
        -----------------    --------------------------------------------------------------------
        save_location        Required string. Path to export the FeatureSet to.
        -----------------    --------------------------------------------------------------------
        out_name             Required string. Name of the saved table.
        -----------------    --------------------------------------------------------------------
        encoding             Optional string. character encoding is used to represent a
                             repertoire of characters by some kind of encoding system. The
                             default is None.
        =================    ====================================================================


        :return: string

        """
        _, file_extension = os.path.splitext(out_name)
        if file_extension.lower() not in ['.csv', '.json'] and \
           HASARCPY == False:
            raise ImportError("ArcPy is required to export a feature class.")
        import sys
        if sys.version_info[0] == 2:
            access = 'wb+'
            kwargs = {}
            if encoding is not None:
                kwargs['encoding'] = encoding
        else:
            access = 'wt+'
            kwargs = {'newline': ''}
            if encoding is not None:
                kwargs['encoding'] = encoding

        if file_extension == ".csv":
            res = os.path.join(save_location, out_name)
            with open(res, access, **kwargs) as csv_file:
                import csv
                csv_writer = csv.writer(csv_file)
                fields = []
                # write the headers to the csv
                for field in self.fields:
                    fields.append(field['name'])
                csv_writer.writerow(fields)

                new_row = []
                # Loop through the results and save each to a row
                for feature in self:
                    new_row = []
                    for field in self.fields:
                        new_row.append(feature.get_value(field['name']))
                    csv_writer.writerow(new_row)
                csv_file.close()
            del csv_file
        elif file_extension == ".json":
            res = os.path.join(save_location, out_name)
            with open(res, access, **kwargs) as writer:
                json.dump(self.value, writer, sort_keys=True, indent=4, ensure_ascii=False)
                writer.flush()
                writer.close()
            del writer
        else:
            temp_dir = tempfile.gettempdir()
            temp_file = os.path.join(temp_dir, "%s.json" % uuid.uuid4().hex)
            with open(temp_file, 'wt', **kwargs) as writer:
                writer.write(self.to_json)
                writer.flush()
                writer.close()
            del writer
            res = json_to_featureclass(json_file=temp_file,
                                       out_fc=os.path.join(save_location, out_name))
            os.remove(temp_file)
        return res

    # ----------------------------------------------------------------------
    @property
    def features(self):
        """gets the features in the FeatureSet"""
        return self._features

    # ----------------------------------------------------------------------
    @property
    def fields(self):
        """gets the fields in the FeatureSet"""
        # todo - build object id field if not found - webmaps need this
        if not self._object_id_field_name:
            obj_field = {'name': 'OBJECTID',
                         'type': 'esriFieldTypeOID',
                         'alias': 'OBJECTID',
                         'sqlType': 'sqlTypeOther'
                         }

            if len(self._fields) > 0:
                field0 = self._fields[0]
                if isinstance(field0, str):
                    self._fields.append('OBJECTID')
                else:
                    self._fields.append(obj_field)
            else:
                self._fields.append(obj_field)

            self._object_id_field_name = obj_field['name']

            counter = 0
            for feat in self.features:
                counter += 1
                feat.fields.append('OBJECTID')
                feat.attributes['OBJECTID'] = counter

        return self._fields

    # ----------------------------------------------------------------------
    @fields.setter
    def fields(self, fields):
        """sets the fields in the FeatureSet"""
        self._fields = fields


class FeatureCollection(Layer):
    """
    FeatureCollection is an object with a layer definition and a feature set.

    It is an in-memory collection of features with rendering information.

    Feature Collections can be stored as Items in the GIS, added as layers to a map or scene,
    passed as inputs to feature analysis tools, and returned as results from feature analysis tools
    if an output name for a feature layer is not specified when calling the tool.
    """

    # noinspection PyMissingConstructor
    def __init__(self, dictdata):
        self._hydrated = True
        self.properties = PropertyMap(dictdata)
        self.layer = self.properties

    @property
    def _lyr_json(self):
        return dict(self.properties)

    @property
    def _lyr_dict(self):
        return dict(self.properties)

    def __str__(self):
        return '<%s>' % type(self).__name__

    def __repr__(self):
        return '<%s>' % type(self).__name__

    def query(self):
        """
        Returns the data in this feature collection as a FeatureSet.
        Filtering by where clause is not supported for feature collections
        """
        if 'layers' in self.properties:
            if 'fields' in self.properties['layers'][0]['layerDefinition']:
                self.properties['layers'][0]['featureSet']['fields'] = \
                    self.properties['layers'][0]['layerDefinition']['fields']

            return FeatureSet.from_dict(self.properties['layers'][0]['featureSet'], )
        else:
            if 'fields' in self.properties['layerDefinition']:
                self.properties['featureSet']['fields'] = self.properties['layerDefinition']['fields']

            return FeatureSet.from_dict(self.properties['featureSet'])

    @staticmethod
    def from_featureset(fset, symbol=None):
        """
        Create a FeatureCollection object from a FeatureSet object.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        fset                   Required arcgis.features.FeatureSet object.
        ------------------     --------------------------------------------------------------------
        symbol                 Optional dict. Specify your symbol as a dictionary. Symbols for points
                               can be picked from http://esri.github.io/arcgis-python-api/tools/symbol.html

                               If not specified, a default symbol will be created.
        ==================     ====================================================================
        :return:
            A FeatureCollection object.
        """
        if not isinstance(fset, FeatureSet):
            raise ValueError

        fset_dict = fset.to_dict()

        # region compose layer definition

        fc_layer_definition = {'geometryType': fset_dict['geometryType'],
                               'fields':fset_dict['fields'],
                               'objectIdField':fset.object_id_field_name,
                               'type':'Feature Layer'}

        if not symbol:
            if fc_layer_definition['geometryType'] == 'esriGeometryPolyline':
                symbol = {"color": [0, 0, 0, 255],
                               "width": 1.33,
                               "type": "esriSLS",
                               "style": "esriSLSSolid"}

            elif fc_layer_definition['geometryType'] in ['esriGeometryPolygon', 'esriGeometryEnvelope']:
                symbol = {"color": [0, 0, 0, 64],
                               "outline": {
                                   "color": [0, 0, 0, 255],
                                   "width": 1.33,
                                   "type": "esriSLS",
                                   "style": "esriSLSSolid"},
                               "type": "esriSFS",
                               "style": "esriSFSSolid"}

            elif fc_layer_definition['geometryType'] in ['esriGeometryPoint', 'esriGeometryMultipoint']:
                symbol = {"angle": 0,
                               "xoffset": 0,
                               "yoffset": 12,
                               "type": "esriPMS",
                               "url": "http://esri.github.io/arcgis-python-api/notebooks/nbimages/pink.png",
                               "contentType": "image/png",
                               "width": 24,
                               "height": 24}

        fc_layer_definition['drawingInfo'] = {'renderer':{'type':'simple',
                                                          'symbol':symbol}
                                              }
        # endregion
        # compose the feature collection dict
        layers_dict = {'featureSet':{'geometryType':fset_dict['geometryType'],
                                   'features':fset_dict['features']},
                    'layerDefinition':fc_layer_definition}

        fc_dict = {'layers':[layers_dict]}

        #create a FC and return
        return FeatureCollection(fc_dict)


from arcgis._impl.common._mixins import PropertyMap
from arcgis.gis import GIS
from arcgis import env
########################################################################
class IdentityProviderManager(object):
    """
    Manages and Updates the SAML identity provider configuration for a given GIS.
    """
    _gis = None
    _portal = None
    _url = None
    _properties = None
    _allowed_keys = None
    #----------------------------------------------------------------------
    def __init__(self, gis=None):
        """Constructor"""
        if gis is None:
            gis = env.active_gis
        isinstance(gis, GIS)
        self._gis = gis
        self._portal = self._gis._portal
        self._url = self._portal.resturl + "portals/self/idp"
        self._allowed_keys = ['groups', 'supportSignedRequest', 'updateProfileAtSignin',
                              'entityId', 'roleId', 'bindingPostUrl', 'certificate',
                              'name', 'logoutUrl', 'id', 'useSHA256', 'encryptionCertificate',
                              'bindingUrl', 'level', 'userCreditAssignment', 'signUpMode',
                              'supportsLogoutRequest', 'encryptionSupported', 'idpMetadataFile']

    #----------------------------------------------------------------------
    def __str__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the properties of the IDP configuration"""
        if self._properties is None:
            params = {'f':'json'}
            res = self._gis._con.get(path=self._url,
                                     params=params)
            try:
                self._properties = PropertyMap(res)
            except:
                self._properties = PropertyMap({})
        return self._properties
    #----------------------------------------------------------------------
    @property
    def configuration(self):
        """
        Gets, updates, or Adds a SAML provider

        ======================  =====================================================================
        **Arguement**           **Value**
        ----------------------  ---------------------------------------------------------------------
        value                   required dictionary.  This property sets, updates or deletes an IDP
                                configuration for a given GIS.

                                To configure an IDP, provide the key/value
                                Example:
                                idp.configuration = {'name' : 'Enterprise IDP',
                                                     'idpMetadataFile' : 'metadata.xml'
                                                    }

                                Once a site has been configured to use IDP, the configuration can be
                                updated by passing in the key/value pair dictionary.
                                Example:

                                idp.configuration = {'name' : 'Acme IDP Login'}

                                To erase an IDP configuration, set the value to None
                                Example:
                                idp.configuration = None

                                Everytime the IDP configuration is updated, the changes can be seen
                                by calling the 'configuration' property and the new results will be
                                returned as a dictionary.
        ======================  =====================================================================

        *Key:Value Dictionary for Argument value*

        ======================  =====================================================================
        **Key**                 **Value**
        ----------------------  ---------------------------------------------------------------------
        bindingPostUrl          Optional string. If the idpMetadataFile isn't specified when an
                                administrator, this parameter is required.  It is federated identity
                                provider post url.
        ----------------------  ---------------------------------------------------------------------
        bindingUrl              Optional string. If the idpMetadataFile isn't specified when an
                                administrator, this parameter is required.  It is federated identity
                                provider url that we have to redirect the user to login to.
        ----------------------  ---------------------------------------------------------------------
        certificate             Optional string. the X509Certificate that needs to be used to
                                validate the SamlResponse from the identity provider.
        ----------------------  ---------------------------------------------------------------------
        encryptionCertificate   Optional string. the X509Certificate that needs to be used to
                                validate the SamlResponse from the identity provider.
        ----------------------  ---------------------------------------------------------------------
        encryptionSupported     Optional bool.  Tells is the SAML provider supports encryption.
        ----------------------  ---------------------------------------------------------------------
        entityId                Optional string.  Name of the entity ID.
        ----------------------  ---------------------------------------------------------------------
        groups                  Optional list. List of group ids that users will be put in on
                                when they signup to join the GIS.
        ----------------------  ---------------------------------------------------------------------
        id                      Optional string. unique identifier of the IDP provider.
        ----------------------  ---------------------------------------------------------------------
        idpMetadataFile         Optional string.  In the case the URL is not accessible, then the
                                same IDP Metadata file can be uploaded.
        ----------------------  ---------------------------------------------------------------------
        level                   Optional integer. Either value 1 or 2. The default level a user will
                                be created as.  The default is 2.
        ----------------------  ---------------------------------------------------------------------
        logoutUrl               Optional string.  The logout SAML url.
        ----------------------  ---------------------------------------------------------------------
        name                    Optional string.  It is the name of the organization's federated
                                identity provider. This is also the name we show up in the Signin
                                page.
        ----------------------  ---------------------------------------------------------------------
        roleId                  Optional string. Default role new users will be.
        ----------------------  ---------------------------------------------------------------------
        signUpMode              Optional string. This is how new users are added to the GIS. There
                                are two modes: Invitation, Automatic
                                Invitation user needs to get an invitation and then signin through
                                federated identity provider.
                                With Automatic all users that signin through the federated identity
                                provider will be added as a user. The privilege/role is set to 'user'
                                Default is Invitation.
        ----------------------  ---------------------------------------------------------------------
        supportSignedRequest    Optional boolean. Determines if signed requests are supported from
                                the provider.
        ----------------------  ---------------------------------------------------------------------
        supportsLogoutRequest   Optional boolean. Determines if logout requests are accepted.
        ----------------------  ---------------------------------------------------------------------
        updateProfileAtSignin   Optional boolean. If True, users have to update the profile.
        ----------------------  ---------------------------------------------------------------------
        useSHA256               Optional boolean. If set to true, SHA256 encryption will be used.
        ----------------------  ---------------------------------------------------------------------
        userCreditAssignment    Optional integer.  Assigns a set number of credits to new users. The
                                default is -1 (infinite).
        ======================  =====================================================================
        """
        return self.properties
    #----------------------------------------------------------------------
    @configuration.setter
    def configuration(self, value):
        """
        Updates or Adds a SAML provider

        ======================  =====================================================================
        **Arguement**           **Value**
        ----------------------  ---------------------------------------------------------------------
        value                   required dictionary.  This property sets, updates or deletes an IDP
                                configuration for a given GIS.

                                To configure an IDP, provide the key/value
                                Example:
                                idp.configuration = {'name' : 'Enterprise IDP',
                                                     'idpMetadataFile' : 'metadata.xml'
                                                    }

                                Once a site has been configured to use IDP, the configuration can be
                                updated by passing in the key/value pair dictionary.
                                Example:

                                idp.configuration = {'name' : 'Acme IDP Login'}

                                To erase an IDP configuration, set the value to None
                                Example:
                                idp.configuration = None

                                Everytime the IDP configuration is updated, the changes can be seen
                                by calling the 'configuration' property and the new results will be
                                returned as a dictionary.
        ======================  =====================================================================

        *Key:Value Dictionary for Argument value*

        ======================  =====================================================================
        **Key**                 **Value**
        ----------------------  ---------------------------------------------------------------------
        bindingPostUrl          Optional string. If the idpMetadataFile isn't specified when an
                                administrator, this parameter is required.  It is federated identity
                                provider post url.
        ----------------------  ---------------------------------------------------------------------
        bindingUrl              Optional string. If the idpMetadataFile isn't specified when an
                                administrator, this parameter is required.  It is federated identity
                                provider url that we have to redirect the user to login to.
        ----------------------  ---------------------------------------------------------------------
        certificate             Optional string. the X509Certificate that needs to be used to
                                validate the SamlResponse from the identity provider.
        ----------------------  ---------------------------------------------------------------------
        encryptionCertificate   Optional string. the X509Certificate that needs to be used to
                                validate the SamlResponse from the identity provider.
        ----------------------  ---------------------------------------------------------------------
        encryptionSupported     Optional bool.  Tells is the SAML provider supports encryption.
        ----------------------  ---------------------------------------------------------------------
        entityId                Optional string.  Name of the entity ID.
        ----------------------  ---------------------------------------------------------------------
        groups                  Optional list. List of group ids that users will be put in on
                                when they signup to join the GIS.
        ----------------------  ---------------------------------------------------------------------
        id                      Optional string. unique identifier of the IDP provider.
        ----------------------  ---------------------------------------------------------------------
        idpMetadataFile         Optional string.  In the case the URL is not accessible, then the
                                same IDP Metadata file can be uploaded.
        ----------------------  ---------------------------------------------------------------------
        level                   Optional integer. Either value 1 or 2. The default level a user will
                                be created as.  The default is 2.
        ----------------------  ---------------------------------------------------------------------
        logoutUrl               Optional string.  The logout SAML url.
        ----------------------  ---------------------------------------------------------------------
        name                    Optional string.  It is the name of the organization's federated
                                identity provider. This is also the name we show up in the Signin
                                page.
        ----------------------  ---------------------------------------------------------------------
        roleId                  Optional string. Default role new users will be.
        ----------------------  ---------------------------------------------------------------------
        signUpMode              Optional string. This is how new users are added to the GIS. There
                                are two modes: Invitation, Automatic
                                Invitation user needs to get an invitation and then signin through
                                federated identity provider.
                                With Automatic all users that signin through the federated identity
                                provider will be added as a user. The privilege/role is set to 'user'
                                Default is Invitation.
        ----------------------  ---------------------------------------------------------------------
        supportSignedRequest    Optional boolean. Determines if signed requests are supported from
                                the provider.
        ----------------------  ---------------------------------------------------------------------
        supportsLogoutRequest   Optional boolean. Determines if logout requests are accepted.
        ----------------------  ---------------------------------------------------------------------
        updateProfileAtSignin   Optional boolean. If True, users have to update the profile.
        ----------------------  ---------------------------------------------------------------------
        useSHA256               Optional boolean. If set to true, SHA256 encryption will be used.
        ----------------------  ---------------------------------------------------------------------
        userCreditAssignment    Optional integer.  Assigns a set number of credits to new users. The
                                default is -1 (infinite).
        ======================  =====================================================================


        """
        if len(dict(self.properties)) == 0 and \
           value is not None:
            self._add(**value)
        elif value is None:
            self._unregister()
        elif len(dict(self.properties)) > 0:
            self._update(**value)
    #----------------------------------------------------------------------
    def _add(self, **kwargs):
        """
        registers the inital idp configuration
        """
        if 'name' not in kwargs:
            import uuid
            kwargs['name'] = uuid.uuid4().hex[:7]
        url = self._url + "/register"
        params = {'f' : 'json'}
        file = kwargs.pop('idpMetadataFile', None)
        params.update(dict(kwargs))
        for k,v in params.items():
            if isinstance(v, list):
                params[k] = ",".join(v)
            else:
                params[k] = v
        if file is not None:
            files = {'idpMetadataFile' : file}
        else:
            files = {}
        res = self._gis._con.post(path=url,
                                  postdata=params,
                                  files=files)
        self._properties = None
        return res
    #----------------------------------------------------------------------
    def _unregister(self):
        """unregisters the current IDP settings"""
        if len(dict(self.properties)) == 0:
            return True
        url = self._url + "/%s/unregister" % self.properties['id']
        params = {'f' : 'json'}
        self._gis._con.post(path=url, postdata=params)
        self._properties = None
    #----------------------------------------------------------------------
    def _update(self, **kwargs):
        """
        updates the idp configuration
        """
        import json
        if 'id' in self._properties:
            url = self._url + "/%s/update" % self.properties['id']
            params = {'f' : 'json'}
            file = kwargs.pop('idpMetadataFile', None)
            params.update(dict(kwargs))
            for k,v in params.items():
                if isinstance(v, list):
                    params[k] = ",".join(v)
                else:
                    params[k] = v
            if file is not None:
                files = {'idpMetadataFile' : file}
            else:
                files = {}
            res = self._gis._con.post(path=url,
                                      postdata=params,
                                      files=files,
                                      try_json=False).replace('",}', '"}')

            self._properties = None
            try:
                return json.loads(res)['success']
            except:
                return True
        return False


"""

Generates Symbol Types

"""
import json




import arcgis
from arcgis.features import FeatureCollection, FeatureSet, SpatialDataFrame
from arcgis.gis import GIS
from arcgis.geometry import _types
###########################################################################
ALLOWED_CMAPS = ['Accent', 'Accent_r', 'Blues', 'Blues_r',
                 'BrBG', 'BrBG_r', 'BuGn', 'BuGn_r', 'BuPu',
             'BuPu_r', 'CMRmap', 'CMRmap_r', 'Dark2',
             'Dark2_r', 'GnBu', 'GnBu_r', 'Greens', 'Greens_r',
             'Greys', 'Greys_r', 'OrRd', 'OrRd_r', 'Oranges',
             'Oranges_r', 'PRGn', 'PRGn_r', 'Paired', 'Paired_r',
             'Pastel1', 'Pastel1_r', 'Pastel2', 'Pastel2_r', 'PiYG',
             'PiYG_r', 'PuBu', 'PuBuGn', 'PuBuGn_r', 'PuBu_r', 'PuOr',
             'PuOr_r', 'PuRd', 'PuRd_r', 'Purples', 'Purples_r',
             'RdBu', 'RdBu_r', 'RdGy', 'RdGy_r', 'RdPu', 'RdPu_r',
             'RdYlBu', 'RdYlBu_r', 'RdYlGn', 'RdYlGn_r', 'Reds',
             'Reds_r', 'Set1', 'Set1_r', 'Set2', 'Set2_r', 'Set3',
             'Set3_r', 'Spectral', 'Spectral_r', 'Wistia', 'Wistia_r',
             'YlGn', 'YlGnBu', 'YlGnBu_r', 'YlGn_r', 'YlOrBr',
             'YlOrBr_r', 'YlOrRd', 'YlOrRd_r', 'afmhot', 'afmhot_r',
             'autumn', 'autumn_r', 'binary', 'binary_r', 'bone',
             'bone_r', 'brg', 'brg_r', 'bwr', 'bwr_r', 'cool', 'cool_r',
             'coolwarm', 'coolwarm_r', 'copper', 'copper_r', 'cubehelix',
             'cubehelix_r', 'flag', 'flag_r', 'gist_earth', 'gist_earth_r',
             'gist_gray', 'gist_gray_r', 'gist_heat', 'gist_heat_r',
             'gist_ncar', 'gist_ncar_r', 'gist_rainbow', 'gist_rainbow_r',
             'gist_stern', 'gist_stern_r', 'gist_yarg', 'gist_yarg_r',
             'gnuplot', 'gnuplot2', 'gnuplot2_r', 'gnuplot_r', 'gray',
             'gray_r', 'hot', 'hot_r', 'hsv', 'hsv_r', 'inferno',
             'inferno_r', 'jet', 'jet_r', 'magma', 'magma_r',
             'nipy_spectral', 'nipy_spectral_r', 'ocean', 'ocean_r',
             'pink', 'pink_r', 'plasma', 'plasma_r', 'prism', 'prism_r',
             'rainbow', 'rainbow_r', 'seismic', 'seismic_r',
             'spring', 'spring_r', 'summer', 'summer_r',
             'terrain', 'terrain_r', 'viridis', 'viridis_r', 'winter',
             'winter_r']
###########################################################################
_CMAP_LOOKUP = dict(zip([c.lower() for c in ALLOWED_CMAPS],
                    ALLOWED_CMAPS))
###########################################################################
_SYMBOL_TYPES = {
    'simple' : 'Simple Point, Line or Polygon Symbol',
    'text' : 'Text Symbol',
    'picture' : 'Picture Marker or Fill Symbol',
    'carto' : 'Cartographic Line Symbol'
}
###########################################################################
LINE_STYLES = {
    's' : 'esriSLSSolid', # default
    '-' : 'esriSLSDash',
    '-.' : 'esriSLSDashDot',
    '-..' : 'esriSLSDashDotDot',
    '.' : 'esriSLSDot',
    '--' : "esriSLSLongDash",
    "--.": "esriSLSLongDashDot",
    'n' :  "esriSLSNull",
    's-' : "esriSLSShortDash",
    's-.' : "esriSLSShortDashDot",
    's-..' : "esriSLSShortDashDotDot",
    's.' : "esriSLSShortDot"
}
###########################################################################
POINT_STYLES = {
    "o" : "esriSMSCircle", #default
    "+" : "esriSMSCross",
    "d" : "esriSMSDiamond",
    "s" : "esriSMSSquare",
    "x" : "esriSMSX",
    #"^" : "esriSMSTriangle" # Does not render in web maps
}
###########################################################################
POLYGON_STYLES = {
    '\\' : "esriSFSBackwardDiagonal",
    "/" : "esriSFSForwardDiagonal",
    "|" : "esriSFSVertical",
    "-" : "esriSFSHorizontal",
    "x" : "esriSFSDiagonalCross",
    "+" : "esriSFSCross",
    "s" : "esriSFSSolid" # default
}
###########################################################################
POLYGON_STYLES_DISPLAY = {
    '\\' : "Backward Diagonal",
    "/" : "Forward Diagonal",
    "|" : "Vertical Bar",
    "-" : "Horizontal Bar",
    "x" : "Diagonal Cross",
    "+" : "Cross",
    "s" : "Solid Fill (default)" # default
}
###########################################################################
POINT_STYLES_DISPLAY = {
    "o" : "Circle (default)", #default
    "+" : "Cross",
    "d" : "Diamond",
    "s" : "Square",
    "x" : "X",
    #"^" : "esriSMSTriangle" # Does not render in web maps
}
###########################################################################
LINE_STYLES_DISPLAY = {
    's' : 'Solid (default)', # default
    '-' : 'Dash',
    '-.' : 'Dash Dot',
    '-..' : 'Dash Dot Dot',
    '.' : 'Dot',
    '--' : "Long Dash",
    "--.": "Long Dash Dot",
    'n' :  "Null",
    's-' : "Short Dash",
    's-.' : "Short Dash Dot",
    's-..' : "Short Dash Dot Dot",
    's.' : "Short Dot"
}
###########################################################################
# Have colormaps separated into categories:
# http://matplotlib.org/examples/color/colormaps_reference.html
cmaps = [('Perceptually Uniform Sequential', ['viridis', 'plasma', 'inferno', 'magma']),
         ('Sequential', [
             'Greys', 'Purples', 'Blues', 'Greens', 'Oranges', 'Reds',
            'YlOrBr', 'YlOrRd', 'OrRd', 'PuRd', 'RdPu', 'BuPu',
            'GnBu', 'PuBu', 'YlGnBu', 'PuBuGn', 'BuGn', 'YlGn']),
         ('Sequential (2)', [
             'binary', 'gist_yarg', 'gist_gray', 'gray', 'bone', 'pink',
             'spring', 'summer', 'autumn', 'winter', 'cool', 'Wistia',
            'hot', 'afmhot', 'gist_heat', 'copper']),
         ('Diverging', [
             'PiYG', 'PRGn', 'BrBG', 'PuOr', 'RdGy', 'RdBu',
             'RdYlBu', 'RdYlGn', 'Spectral', 'coolwarm', 'bwr', 'seismic']),
         ('Qualitative', [
             'Pastel1', 'Pastel2', 'Paired', 'Accent',
             'Dark2', 'Set1', 'Set2', 'Set3',
            'tab10', 'tab20', 'tab20b', 'tab20c']),
         ('Miscellaneous', [
             'flag', 'prism', 'ocean', 'gist_earth', 'terrain', 'gist_stern',
             'gnuplot', 'gnuplot2', 'CMRmap', 'cubehelix', 'brg', 'hsv',
            'gist_rainbow', 'rainbow', 'jet', 'nipy_spectral', 'gist_ncar'])]
###########################################################################
def _cmap2rgb(colors, step, alpha=1):
    """converts a color map to RGBA list"""
    from matplotlib import cm
    t = getattr(cm, colors)(step, bytes=True)
    t = [int(i) for i in t]
    t[-1] = alpha * 255
    return t
###########################################################################
def display_colormaps(colors=None):
    """
    Displays a visual colormaps in order to assist users in selecting a color scheme for
    the data they wish to display on a map, or in a web map.

    This is a variation of http://matplotlib.org/examples/color/colormaps_reference.html

    """
    import numpy as np
    import matplotlib.pyplot as plt

    if colors is None:
        cmaps =  [('All Color Maps', ALLOWED_CMAPS)]
    elif isinstance(colors, (str)):
        cmaps = [('Selected Color Map', [_CMAP_LOOKUP[colors.lower()]])]
    elif isinstance(colors, (tuple, list)):
        colors = [_CMAP_LOOKUP[color.lower()] for color in colors]
        cmaps = [('Selected Color Map', colors)]
    nrows = max(len(cmap_list) for cmap_category, cmap_list in cmaps)
    gradient = np.linspace(0, 1, 256)
    gradient = np.vstack((gradient, gradient))
    def plot_color_gradients(cmap_category, cmap_list, nrows):
        """creates the entry for the plot"""
        if nrows == 1:
            nrows = 2
        height = .2 * nrows
        fig, axes = plt.subplots(nrows=nrows, figsize=(6, height))
        fig.subplots_adjust(top=0.95, bottom=0.01, left=0.2, right=0.99)
        axes[0].set_title(cmap_category, fontsize=12)

        for ax, name in zip(axes, cmap_list):
            ax.imshow(gradient, aspect='auto', cmap=plt.get_cmap(name))
            pos = list(ax.get_position().bounds)
            x_text = pos[0] - 0.01
            y_text = pos[1] + pos[3]/2.
            fig.text(x_text, y_text, name, va='center', ha='right', fontsize=10)

        # Turn off *all* ticks & spines, not just the ones with colormaps.
        for ax in axes:
            ax.set_axis_off()


    for cmap_category, cmap_list in cmaps:
        plot_color_gradients(cmap_category, cmap_list, nrows)

    plt.show()
###########################################################################
def show_styles(geometry_type):
    """
    Returns the available styles for a given geometry type as a Pandas dataframe
    """
    import pandas as pd
    if geometry_type.lower() in ['point', 'multipoint']:
        c = POINT_STYLES_DISPLAY
    elif geometry_type.lower() in ['polyline', 'line']:
        c = LINE_STYLES_DISPLAY
    elif geometry_type.lower() in ['polygon']:
        c = POLYGON_STYLES_DISPLAY
    else:
        raise Exception("Invalid geometry type")

    d = [(k,v) for k,v in c.items()]
    return pd.DataFrame(data=d, columns=['MARKER', 'ESRI_STYLE'])
###########################################################################
def create_symbol(geometry_type,
                  symbol_type=None,
                  symbol_style=None,
                  colors=None,
                  **kwargs):
    """
    Generates a Symbol from a given set of Paramters

    This method creates either a Picture, Text, Cartographic, or Simple Symbol
    based on a given set of parameters.  Each symbol type has a specific set of
    parameters that are excepted.  There is a simplified input definition similar
    to matplotlib.


    =======================  =========================================================
    **Required Argument**    **Description**
    -----------------------  ---------------------------------------------------------
    geometry_type            required string.  This can be point, line, polygon, or
                             multipoint.  It helps to ensure that the symbol created
                             will be supported by that geometry type.
    =======================  =========================================================


    =======================  =========================================================
    **Optional Argument**    **Description**
    -----------------------  ---------------------------------------------------------
    symbol_type              optional string. This is the type of symbol the user
                             needs to create.  Valid inputs are: simple, picture, text,
                             or carto.  The default is simple.
    -----------------------  ---------------------------------------------------------
    symbol_style             optional string. This is the symbology used by the
                             geometry.  For example 's' for a Line geometry is a solid
                             line. And '-' is a dash line.

                             **Point Symbols**

                             + 'o' - Circle (default)
                             + '+' - Cross
                             + 'D' - Diamond
                             + 's' - Square
                             + 'x' - X

                             **Polyline Symbols**

                             + 's' - Solid (default)
                             + '-' - Dash
                             + '-.' - Dash Dot
                             + '-..' - Dash Dot Dot
                             + '.' - Dot
                             + '--' - Long Dash
                             + '--.' - Long Dash Dot
                             + 'n' - Null
                             + 's-' - Short Dash
                             + 's-.' - Short Dash Dot
                             + 's-..' - Short Dash Dot Dot
                             + 's.' - Short Dot

                             **Polygon Symbols**

                             + 's' - Solid Fill (default)
                             + '\' - Backward Diagonal
                             + '/' - Forward Diagonal
                             + '|' - Vertical Bar
                             + '-' - Horizontal Bar
                             + 'x' - Diagonal Cross
                             + '+' - Cross
    -----------------------  ---------------------------------------------------------
    colors                     optional string or list.  This is the color scheme a user
                             can provide if the exact color is not needed, or a user
                             can provide a list with the color defined as:
                             [red, green blue, alpha]. The values red, green, blue are
                             from 0-255 and alpha is a float value from 0 - 1.
                             The default value is 'jet' color scheme.
    -----------------------  ---------------------------------------------------------
    cstep                    optional integer.  If provided, its the color location on
                             the color scheme.
    =======================  =========================================================

    **Simple Symbols**

    This is a list of optional parameters that can be given for point, line or
    polygon geometries.

    ====================  =========================================================
    **Argument**          **Description**
    --------------------  ---------------------------------------------------------
    marker_size           optional float.  Numeric size of the symbol given in
                          points.
    --------------------  ---------------------------------------------------------
    marker_angle          optional float. Numeric value used to rotate the symbol.
                          The symbol is rotated counter-clockwise. For example,
                          The following, angle=-30, in will create a symbol rotated
                          -30 degrees counter-clockwise; that is, 30 degrees
                          clockwise.
    --------------------  ---------------------------------------------------------
    marker_xoffset        Numeric value indicating the offset on the x-axis in points.
    --------------------  ---------------------------------------------------------
    marker_yoffset        Numeric value indicating the offset on the y-axis in points.
    --------------------  ---------------------------------------------------------
    line_width            optional float. Numeric value indicating the width of the line in points
    --------------------  ---------------------------------------------------------
    outline_style         Optional string. For polygon point, and line geometries , a
                          customized outline type can be provided.

                          Allowed Styles:

                          + 's' - Solid (default)
                          + '-' - Dash
                          + '-.' - Dash Dot
                          + '-..' - Dash Dot Dot
                          + '.' - Dot
                          + '--' - Long Dash
                          + '--.' - Long Dash Dot
                          + 'n' - Null
                          + 's-' - Short Dash
                          + 's-.' - Short Dash Dot
                          + 's-..' - Short Dash Dot Dot
                          + 's.' - Short Dot
    --------------------  ---------------------------------------------------------
    outline_color         optional string or list.  This is the same color as the
                          colors property, but specifically applies to the outline_color.
    ====================  =========================================================

    **Picture Symbol**

    This type of symbol only applies to Points, MultiPoints and Polygons.

    ====================  =========================================================
    **Argument**          **Description**
    --------------------  ---------------------------------------------------------
    marker_angle          Numeric value that defines the number of degrees ranging
                          from 0-360, that a marker symbol is rotated. The rotation
                          is from East in a counter-clockwise direction where East
                          is the 0 axis.
    --------------------  ---------------------------------------------------------
    marker_xoffset        Numeric value indicating the offset on the x-axis in points.
    --------------------  ---------------------------------------------------------
    marker_yoffset        Numeric value indicating the offset on the y-axis in points.
    --------------------  ---------------------------------------------------------
    height                Numeric value used if needing to resize the symbol. Specify a value in points. If images are to be displayed in their original size, leave this blank.
    --------------------  ---------------------------------------------------------
    width                 Numeric value used if needing to resize the symbol. Specify a value in points. If images are to be displayed in their original size, leave this blank.
    --------------------  ---------------------------------------------------------
    url                   String value indicating the URL of the image. The URL should be relative if working with static layers. A full URL should be used for map service dynamic layers. A relative URL can be dereferenced by accessing the map layer image resource or the feature layer image resource.
    --------------------  ---------------------------------------------------------
    image_data            String value indicating the base64 encoded data.
    --------------------  ---------------------------------------------------------
    xscale                Numeric value indicating the scale factor in x direction.
    --------------------  ---------------------------------------------------------
    yscale                Numeric value indicating the scale factor in y direction.
    --------------------  ---------------------------------------------------------
    outline_color         optional string or list.  This is the same color as the
                          colors property, but specifically applies to the outline_color.
    --------------------  ---------------------------------------------------------
    outline_style         Optional string. For polygon point, and line geometries , a
                          customized outline type can be provided.

                          Allowed Styles:

                          + 's' - Solid (default)
                          + '-' - Dash
                          + '-.' - Dash Dot
                          + '-..' - Dash Dot Dot
                          + '.' - Dot
                          + '--' - Long Dash
                          + '--.' - Long Dash Dot
                          + 'n' - Null
                          + 's-' - Short Dash
                          + 's-.' - Short Dash Dot
                          + 's-..' - Short Dash Dot Dot
                          + 's.' - Short Dot
    --------------------  ---------------------------------------------------------
    outline_color         optional string or list.  This is the same color as the
                          colors property, but specifically applies to the outline_color.
    --------------------  ---------------------------------------------------------
    line_width            optional float. Numeric value indicating the width of the line in points
    ====================  =========================================================

    **Text Symbol**

    This type of symbol only applies to Points, MultiPoints and Polygons.

    ====================  =========================================================
    **Argument**          **Description**
    --------------------  ---------------------------------------------------------
    font_decoration       The text decoration. Must be one of the following values:
                          - line-through
                          - underline
                          - none
    --------------------  ---------------------------------------------------------
    font_family           Optional string. The font family.
    --------------------  ---------------------------------------------------------
    font_size             Optional float. The font size in points.
    --------------------  ---------------------------------------------------------
    font_style            Optional string. The text style.
                          - italic
                          - normal
                          - oblique
    --------------------  ---------------------------------------------------------
    font_weight           Optional string. The text weight.
                          Must be one of the following values:
                          - bold
                          - bolder
                          - lighter
                          - normal
    --------------------  ---------------------------------------------------------
    background_color      optional string/list. Background color is represented as
                          a four-element array or string of a color map.
    --------------------  ---------------------------------------------------------
    halo_color            Optional string/list. Color of the halo around the text.
                          The default is None.
    --------------------  ---------------------------------------------------------
    halo_size             Optional integer/float. The point size of a halo around
                          the text symbol.
    --------------------  ---------------------------------------------------------
    horizontal_alignment  optional string. One of the following string values
                          representing the horizontal alignment of the text.
                          Must be one of the following values:
                          - left
                          - right
                          - center
                          - justify
    --------------------  ---------------------------------------------------------
    kerning               optional boolean. Boolean value indicating whether to
                          adjust the spacing between characters in the text string.
    --------------------  ---------------------------------------------------------
    line_color            optional string/list. Outline color is represented as
                          a four-element array or string of a color map.
    --------------------  ---------------------------------------------------------
    line_width            optional integer/float. Outline size.
    --------------------  ---------------------------------------------------------
    marker_angle          optional int. A numeric value that defines the number of
                          degrees (0 to 360) that a text symbol is rotated. The
                          rotation is from East in a counter-clockwise direction
                          where East is the 0 axis.
    --------------------  ---------------------------------------------------------
    marker_xoffset        optional int/float.Numeric value indicating the offset
                          on the x-axis in points.
    --------------------  ---------------------------------------------------------
    marker_yoffset        optional int/float.Numeric value indicating the offset
                          on the x-axis in points.
    --------------------  ---------------------------------------------------------
    right_to_left         optional boolean. Set to true if using Hebrew or Arabic
                          fonts.
    --------------------  ---------------------------------------------------------
    rotated               optional boolean. Boolean value indicating whether every
                          character in the text string is rotated.
    --------------------  ---------------------------------------------------------
    text                  Required string.  Text Value to display next to geometry.
    --------------------  ---------------------------------------------------------
    vertical_alignment    Optional string. One of the following string values
                          representing the vertical alignment of the text.
                          Must be one of the following values:
                          - top
                          - bottom
                          - middle
                          - baseline
    ====================  =========================================================

    **Cartographic Symbol**

    This type of symbol only applies to line geometries.

    ====================  =========================================================
    **Argument**          **Description**
    --------------------  ---------------------------------------------------------
    line_width            optional float. Numeric value indicating the width of the line in points
    --------------------  ---------------------------------------------------------
    cap                   Optional string.  The cap style.
    --------------------  ---------------------------------------------------------
    join                  Optional string. The join style.
    --------------------  ---------------------------------------------------------
    miter_limit           Optional string. Size threshold for showing mitered line joins.
    ====================  =========================================================

    :returns: Dictionary

    """
    import numpy as np
    import matplotlib.pyplot as plt

    alpha = kwargs.pop('alpha', 1)
    symbol = kwargs.pop('symbol', None)
    if symbol_type is None:
        symbol_type = "simple"

    gtype = geometry_type.upper()
    if colors is None:
        colors = 'jet'
    cstep = kwargs.pop('cstep', 1)
    renderer_type = "simple"

    marker_size = kwargs.pop('marker_size', 8)
    marker_angle = kwargs.pop('marker_angle', 0)
    marker_xoffset = kwargs.pop('marker_xoffset', 0)
    marker_yoffset = kwargs.pop('marker_yoffset', 0)

    line_width = kwargs.pop('line_width', 2)

    outline_style = LINE_STYLES[kwargs.pop('outline_style', "s")]
    outline_color = kwargs.pop('outline_color', colors)

    if isinstance(colors, str):
        colors = list(
            _cmap2rgb(colors=colors, step=cstep,alpha=alpha))
    elif isinstance(colors, (list, tuple)) and \
         len(colors) == 4:
        colors = colors
    else:
        colors = list(_cmap2rgb(
            colors='jet',
            step=cstep,
            alpha=alpha))
    if isinstance(outline_color, str):
        outline_cmap = list(_cmap2rgb(colors=outline_color,
                                      step=int(np.random.randint(0,255)),
                                      alpha=alpha))
    elif isinstance(outline_color, (tuple, list)) and \
         len(outline_color) == 4:
        outline_cmap = outline_color
    else:
        outline_cmap = list(_cmap2rgb(colors='jet',
                                      step=int(np.random.randint(0,255)),
                                      alpha=alpha))

    if symbol is not None:
        return symbol

    if symbol_type.lower() == 'simple': # Default Simple Symbol

        if symbol_style is None:
            if gtype == "POINT":
                symbol_type = "esriSMS"
                symbol_style = POINT_STYLES["o"]
            elif gtype in ["POLYLINE", 'LINE']:
                symbol_type = "esriSLS"
                symbol_style = LINE_STYLES['s']
            elif gtype in ['POLYGON']:
                symbol_type = "esriSFS"
                symbol_style = POLYGON_STYLES['s']
            else:
                raise Exception("Invalid geometry types only points, lines, and polygons can be plotted")
        else:
            if gtype == "POINT":
                symbol_type = "esriSMS"
                symbol_style = POINT_STYLES[symbol_style]
            elif gtype in ["POLYLINE", 'LINE']:
                symbol_type = "esriSLS"
                symbol_style = LINE_STYLES[symbol_style]
            elif gtype in ['POLYGON']:
                symbol_type = "esriSFS"
                symbol_style = POLYGON_STYLES[symbol_style]
            else:
                raise Exception("Invalid geometry types only points, lines, and polygons can be plotted")
        # build the symbol
        symbol = { }
        symbol["type"] = symbol_type
        if isinstance(colors, str):
            colors = list(
                _cmap2rgb(colors=colors, step=cstep,alpha=alpha))
        elif isinstance(colors, (list, tuple)) and \
             len(colors) == 4:
            colors = colors
        else:
            colors = list(_cmap2rgb(
                colors='jet',
                step=cstep,
                alpha=alpha))
        if isinstance(outline_color, str):
            outline_cmap = list(_cmap2rgb(colors=outline_color,
                                          step=int(np.random.randint(0,255)),
                                          alpha=alpha))
        elif isinstance(outline_color, (tuple, list)) and \
             len(outline_color) == 4:
            outline_cmap = outline_color
        else:
            outline_cmap = list(_cmap2rgb(colors='jet',
                                          step=int(np.random.randint(0,255)),
                                          alpha=alpha))
        if gtype == "POINT":
            symbol["style"] = symbol_style
            symbol["color"] = colors
            symbol["size"] = marker_size
            symbol["angle"] = marker_angle
            symbol["xoffset"] = marker_xoffset
            symbol["yoffset"] = marker_yoffset
            symbol["outline"] = {
                'style': outline_style,
                'type': 'esriSLS',
                "color" : outline_cmap,
                "width" : line_width
            }
        elif gtype in ['LINE', 'POLYLINE']:
            symbol["style"] = symbol_style
            symbol["color"] = colors
            symbol["width"] = line_width
        elif gtype in ['POLYGON']:
            symbol["style"] = symbol_style
            symbol["color"] = colors
            symbol['outline'] = {
                'style': outline_style,
                'type': 'esriSLS',
                "color" : outline_cmap,
                "width" : line_width
            }
        return symbol
    elif symbol_type == 'picture' and \
         geometry_type.lower() in ['multipoint',
                                   'point',
                                   'polygon']:
        if geometry_type.lower() in ['point', 'multipoint']:
            symbol = {
                "type" : "esriPMS",
                "angle" : marker_angle,
                "xoffset" : marker_xoffset,
                "yoffset" : marker_yoffset,
                "width" : kwargs.pop('width', 10),
                "height" : kwargs.pop('height', 10),
                "url" : kwargs.pop('url', ""),
                "imageData" : kwargs.pop('image_data', "")
            }
            return symbol
        elif geometry_type.lower() == "polygon":
            symbol = {
                "type" : "esriPFS",
                "color" : colors,
                "xoffset" : marker_xoffset,
                "yoffset" : marker_yoffset,
                "xscale" : kwargs.pop('xscale', None),
                "yscale" : kwargs.pop('yscale', None),
                "width" : kwargs.pop('width', 10),
                "height" : kwargs.pop('height', 10),
                "angle" : marker_angle,
                "url" : kwargs.pop('url', ""),
                "imageData" : kwargs.pop('image_data', None),
                "outline" : {
                   "color" : outline_cmap,
                   "width" : line_width,
                   "type" : "esriSLS",
                   "style" : outline_style
                }
            }
            return symbol
    elif symbol_type.lower() == 'picture' and \
         geometry_type.lower() in ['line']:
        raise Exception("Line geometries do not support picture symbology")
    elif symbol_type.lower() == 'text' and \
         geometry_type.lower() in ['point']:
        symbol = {
            "color": colors,
            "type": "esriTS",
            "horizontalAlignment": kwargs.pop("horizontal_alignment", 'center'),
            'verticalAlignment' : kwargs.pop('vertical_alignment', 'middle'),
            'rightToLeft' : kwargs.pop('right_to_left', False),
            "backgroundColor" : kwargs.pop('background_color', colors),
            "borderLineColor" : outline_cmap,
            "borderLineSize" : line_width,
            "haloColor" : kwargs.pop('halo_color', None),
            "haloSize" : kwargs.pop('halo_size', None),
            "angle": marker_angle,
            "xoffset": marker_xoffset,
            "yoffset": marker_yoffset,
            "text": kwargs.pop("text" , "N/A"),
            "rotated": kwargs.pop('rotated', False),
            "kerning": kwargs.pop('kerning', True),
            "font": {
                "size": kwargs.pop("font_size", 10),
                "style": kwargs.pop("font_style", "normal"),
                "weight": kwargs.pop("font_weight", "normal"),
                "family": kwargs.pop("font_familty", "serif"),
                'decoration' : kwargs.pop('font_decoration', 'none')
            }
        }
        return symbol
    elif symbol_type.lower() == 'text' and \
         geometry_type.lower() not in ['point', 'multipoint']:
        raise Exception("Text symbology is only supported by points and multipoints")
    elif symbol_type.lower() == 'carto' and \
         geometry_type in ['line', 'polyline']:
        cap = kwargs.pop('cap', "esriLCButt")
        join = kwargs.pop('join', 'esriLJSMiter')
        miter_limit = kwargs.pop('miter_limit', 1)
        symbol = {
            "color" : colors,
            'width' : line_width,
            'type' : "esriCLS",
            "cap" : cap,
            "join" : join,
            "miter_limit" : miter_limit
        }
    elif symbol_type.lower() == 'carto' and \
         geometry_type not in ['line']:
        raise Exception("Cartographic symbology is only supported for line geomtries")
    else:
        raise Exception("Invalid symbol and geometry type")
    return symbol



from .._layer import ImageryLayer
from arcgis.gis import Item
import numbers

def _raster_input(raster, raster2=None):
    if raster2 is not None:
        if isinstance(raster2, ImageryLayer) and isinstance(raster, ImageryLayer):
            layer = raster2
            raster_ra = _get_raster_ra(raster2)
            if raster2._fn is not None:
                if raster._url == raster2._url:
                    raster2 = raster2._fn
                else:
                    raster2 = _replace_raster_url(raster2._fn, raster2._url)
            else:
                if raster2._url == raster._url:
                    oids = raster2.filtered_rasters()
                    if oids is None:
                        raster2 = '$$'
                    elif len(oids) == 1:
                        raster2 = '$' + str(oids[0])
                    else:
                        raster2 = ['$' + str(x) for x in oids]
                else:
                    raster2 = raster2._url
        elif isinstance(raster2, ImageryLayer) and not isinstance(raster, ImageryLayer):
            layer = raster2
            raster_ra = _get_raster_ra(raster2)
            raster2 = _get_raster(raster2)

        elif isinstance(raster2, list):
            mix_and_match = False # mixing rasters from two image services
            # try:
            #     r0 = raster2[0]
            #     r1 = raster2[1]
            #     if r0._fn is None and r1._fn is None and r0._url != r1._url:
            #         mix_and_match = True
            # except:
            #     pass

            for r in raster2: # layer is first non numeric raster in list
                if not isinstance(r, numbers.Number):
                    layer = r
                    break

            for r in raster2:
                if not isinstance(r, numbers.Number):
                    if r._url != layer._url:
                        mix_and_match = True

            raster_ra = [_get_raster_ra(r) for r in raster2]
            if mix_and_match:
                raster2 = [_get_raster_url(r, layer) for r in raster2]
            else:
                raster2 = [_get_raster(r) for r in raster2]
        else: # secondinput maybe scalar for arithmetic functions, or a chained raster fn
            layer = None
            # raster = raster
            raster_ra = raster2
        return layer, raster2, raster_ra

    if isinstance(raster, ImageryLayer):
        layer = raster
        raster_ra = _get_raster_ra(raster)
        raster = _get_raster(raster)

    elif isinstance(raster, list):
        mix_and_match = False # mixing rasters from two image services
        # try:
        #     r0 = raster[0]
        #     r1 = raster[1]
        #     if r0._fn is None and r1._fn is None and r0._url != r1._url:
        #         mix_and_match = True
        # except:
        #     pass

        for r in raster: # layer is first non numeric raster in list
            if not isinstance(r, numbers.Number):
                layer = r
                break

        for r in raster:
            if not isinstance(r, numbers.Number):
                if r._url != layer._url:
                    mix_and_match = True

        raster_ra = [_get_raster_ra(r) for r in raster]
        if mix_and_match:
            raster = [_get_raster_url(r, layer) for r in raster]
        else:
            raster = [_get_raster(r) for r in raster]
    else: # maybe scalar for arithmetic functions, or a chained raster fn
        layer = None
        # raster = raster
        raster_ra = raster

    return layer, raster, raster_ra

def _get_raster(raster):
    if isinstance(raster, ImageryLayer):
        if raster._fn is not None:
            raster = raster._fn
        else:
            oids = raster.filtered_rasters()
            if oids is None:
                raster = '$$'
            elif len(oids) == 1:
                raster = '$' + str(oids[0])
            else:
                raster = ['$' + str(x) for x in oids]
    return raster


def _replace_raster_url(obj, url=None):
    # replace all "Raster" : '$$' with url
    if isinstance(obj, dict):
        value = {k: _replace_raster_url(v, url)
                 for k, v in obj.items()}
    elif isinstance(obj, list):
        value = [_replace_raster_url(elem, url)
                 for elem in obj]
    else:
        value = obj

    if value == '$$':
        return url
    elif isinstance(value, str) and len(value) > 0 and value[0] == '$':
        return url + '/' + value.replace('$', '')
    else:
        return value


def _get_raster_url(raster, layer):
    if isinstance(raster, ImageryLayer):
        if raster._fn is not None:
            if raster._url == layer._url:
                raster = raster._fn
            else:
                raster = _replace_raster_url(raster._fn, raster._url)

        else:
            if raster._url == layer._url:
                raster = '$$'
            else:
                raster = raster._url

            # oids = raster.filtered_rasters()
            # if oids is None:
            #     raster = '$$'
            # elif len(oids) == 1:
            #     raster = '$' + str(oids[0])
            # else:
            #     raster = ['$' + str(x) for x in oids]
    return raster


def _get_raster_ra(raster):
    if isinstance(raster, ImageryLayer):
        if raster._fnra is not None:
            raster_ra = raster._fnra
        else:
            raster_ra = raster._url


            #if raster._mosaic_rule is not None:
            #    raster_ra['mosaicRule'] = raster._mosaic_rule
    elif isinstance(raster, Item):
        raise RuntimeError('Item not supported as input. Use ImageryLayer - e.g. item.layers[0]')
        #raster_ra = {
        #    'itemId': raster.itemid
        #}
    else:
        raster_ra = raster

    return raster_ra




"""
These tools help you identify, quantify, and visualize spatial patterns in your data.

calculate_density takes known quantities of some phenomenon and spreads these quantities across the map.
find_hot_spots identifies statistically significant clustering in the spatial pattern of your data.
"""
import json as _json
from datetime import datetime as _datetime
import logging as _logging
import arcgis as _arcgis
from arcgis.features import FeatureSet as _FeatureSet
from arcgis.geoprocessing._support import _execute_gp_tool
from arcgis.geoprocessing import DataFile
from ._util import _id_generator, _feature_input, _set_context, _create_output_service

_log=_logging.getLogger(__name__)

_use_async=True

def calculate_density(
    input_layer,
    fields=None,
    weight="""Uniform""",
    bin_type="""Square""",
    bin_size=None,
    bin_size_unit=None,
    time_step_interval=None,
    time_step_interval_unit=None,
    time_step_repeat_interval=None,
    time_step_repeat_interval_unit=None,
    time_step_reference=None,
    radius=None,
    radius_unit=None,
    area_units="""SquareKilometers""",
    output_name=None,
    gis=None):
    """




Parameters:

   input_layer: Input Points (Feature layer). Required parameter.

   fields: Population Field (str). Optional parameter.

   weight: Weight (str). Required parameter.
      Choice list:['Uniform', 'Kernel']

   bin_type: Output Bin Type (str). Required parameter.
      Choice list:['Square', 'Hexagon']

   bin_size: Output Bin Size (float). Required parameter.

   bin_size_unit: Output Bin Size Unit (str). Required parameter.
      Choice list:['Feet', 'Yards', 'Miles', 'Meters', 'Kilometers', 'NauticalMiles']

   time_step_interval: Time Step Interval (int). Optional parameter.

   time_step_interval_unit: Time Step Interval Unit (str). Optional parameter.
      Choice list:['Years', 'Months', 'Weeks', 'Days', 'Hours', 'Minutes', 'Seconds', 'Milliseconds']

   time_step_repeat_interval: Time Step Repeat Interval (int). Optional parameter.

   time_step_repeat_interval_unit: Time Step Repeat Interval Unit (str). Optional parameter.
      Choice list:['Years', 'Months', 'Weeks', 'Days', 'Hours', 'Minutes', 'Seconds', 'Milliseconds']

   time_step_reference: Time Step Reference (_datetime). Optional parameter.

   radius: Radius (float). Required parameter.

   radius_unit: Radius Unit (str). Required parameter.
      Choice list:['Feet', 'Yards', 'Miles', 'Meters', 'Kilometers', 'NauticalMiles']

   area_units: Area Unit Scale Factor (str). Optional parameter.
      Choice list:['SquareMeters', 'SquareKilometers', 'Hectares', 'SquareFeet', 'SquareYards', 'SquareMiles', 'Acres']

   output_name: Output Features Name (str). Required parameter.

   gis: Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


Returns:
   output - Output Features as a feature layer collection item


    """
    kwargs=locals()

    gis=_arcgis.env.active_gis if gis is None else gis
    url=gis.properties.helperServices.geoanalytics.url

    params={}
    for key, value in kwargs.items():
        if value is not None:
            params[key]=value

    if output_name is None:
        output_service_name='Calculate Density Analysis_' + _id_generator()
        output_name=output_service_name.replace(' ', '_')
    else:
        output_service_name=output_name.replace(' ', '_')

    output_service=_create_output_service(gis, output_name, output_service_name, 'Calculate Density')

    params['output_name']=_json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db={
        "input_layer": (_FeatureSet, "inputLayer"),
        "fields": (str, "fields"),
        "weight": (str, "weight"),
        "bin_type": (str, "binType"),
        "bin_size": (float, "binSize"),
        "bin_size_unit": (str, "binSizeUnit"),
        "time_step_interval": (int, "timeStepInterval"),
        "time_step_interval_unit": (str, "timeStepIntervalUnit"),
        "time_step_repeat_interval": (int, "timeStepRepeatInterval"),
        "time_step_repeat_interval_unit": (str, "timeStepRepeatIntervalUnit"),
        "time_step_reference": (_datetime, "timeStepReference"),
        "radius": (float, "radius"),
        "radius_unit": (str, "radiusUnit"),
        "area_units": (str, "areaUnits"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }
    return_values=[
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]

    try:
        _execute_gp_tool(gis, "CalculateDensity", params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise


calculate_density.__annotations__={
    'fields': str,
    'weight': str,
    'bin_type': str,
    'bin_size': float,
    'bin_size_unit': str,
    'time_step_interval': int,
    'time_step_interval_unit': str,
    'time_step_repeat_interval': int,
    'time_step_repeat_interval_unit': str,
    'time_step_reference': _datetime,
    'radius': float,
    'radius_unit': str,
    'area_units': str,
    'output_name': str}


def find_hot_spots(
    point_layer,
    bin_size=5,
    bin_size_unit="Miles",
    neighborhood_distance=5,
    neighborhood_distance_unit="Miles",
    time_step_interval=None,
    time_step_interval_unit=None,
    time_step_alignment=None,
    time_step_reference=None,
    cell_size=None,
    cell_size_units=None,
    shape_type=None,
    output_name=None,
    gis=None):
    """



    Parameters:

       point_layer: Input Points (FeatureSet). Required parameter.

       bin_size: Bin Size (float). Optional parameter.

       bin_size_unit: Bin Size Unit (str). Optional parameter.
          Choice list:['Feet', 'Yards', 'Miles', 'Meters', 'Kilometers', 'NauticalMiles']

       neighborhood_distance: Neighborhood Distance (float). Optional parameter.

       neighborhood_distance_unit: Neighborhood Distance Unit (str). Optional parameter.
          Choice list:['Feet', 'Yards', 'Miles', 'Meters', 'Kilometers', 'NauticalMiles']

       time_step_interval: Time Step Interval (int). Optional parameter.

       time_step_interval_unit: Time Step Interval Unit (str). Optional parameter.
          Choice list:['Years', 'Months', 'Weeks', 'Days', 'Hours', 'Minutes', 'Seconds', 'Milliseconds']

       time_step_alignment: Time Step Alignment (str). Optional parameter.
          Choice list:['EndTime', 'StartTime', 'ReferenceTime']

       time_step_reference: Time Step Reference (_datetime). Optional parameter.

       cell_size: optional integer determining the grid size.

       cell_size_units: optional string. The unit of the cell size.

       shape_type: optional string.  The cell shape.

       output_name: Output Features Name (str). Optional parameter.

       gis: Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


    Returns:
       output - Output Features as a feature layer collection item


    """
    kwargs=locals()

    gis=_arcgis.env.active_gis if gis is None else gis
    url=gis.properties.helperServices.geoanalytics.url

    params={}
    for key, value in kwargs.items():
        if value is not None:
            params[key]=value

    if output_name is None:
        output_service_name='Hotspot Analysis_' + _id_generator()
        output_name=output_service_name.replace(' ', '_')
    else:
        output_service_name=output_name.replace(' ', '_')

    output_service=_create_output_service(gis, output_name, output_service_name, 'Find Hotspots')

    params['output_name']=_json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db={
        "point_layer": (_FeatureSet, "pointLayer"),
        "bin_size": (float, "binSize"),
        "bin_size_unit": (str, "binSizeUnit"),
        "neighborhood_distance": (float, "neighborhoodDistance"),
        "neighborhood_distance_unit": (str, "neighborhoodDistanceUnit"),
        "time_step_interval": (int, "timeStepInterval"),
        "time_step_interval_unit": (str, "timeStepIntervalUnit"),
        "time_step_alignment": (str, "timeStepAlignment"),
        "time_step_reference": (_datetime, "timeStepReference"),
        "cell_size" : (int, "cellSize"),
        "cell_size_units": (str, "cellSizeUnits"),
        "shape_type" : (str, "shapeType"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }
    return_values=[
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]

    try:
        _execute_gp_tool(gis, "FindHotSpots", params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise

find_hot_spots.__annotations__={
    'bin_size': float,
    'bin_size_unit': str,
    'neighborhood_distance': float,
    'neighborhood_distance_unit': str,
    'time_step_interval': int,
    'time_step_interval_unit': str,
    'time_step_alignment': str,
    'time_step_reference': _datetime,
    'output_name': str}


def create_space_time_cube(point_layer: _FeatureSet,
                           bin_size: float,
                           bin_size_unit: str,
                           time_step_interval: int,
                           time_step_interval_unit: str,
                           time_step_alignment: str=None,
                           time_step_reference: _datetime=None,
                           summary_fields: str=None,
                           output_name: str=None,
                           context: str=None,
                           gis=None) -> DataFile:
    """
    Summarizes a set of points into a netCDF data structure by aggregating them into space-time bins. Within each bin,
    the points are counted and specified attributes are aggregated. For all bin locations, the trend for counts and
    summary field values are evaluated.

    Parameters:

       point_layer: Input Features (FeatureSet). Required parameter.

       bin_size: Distance Interval (float). Required parameter.

       bin_size_unit: Distance Interval Unit (str). Required parameter.
          Choice list:['Feet', 'Yards', 'Miles', 'Meters', 'Kilometers', 'NauticalMiles']

       time_step_interval: Time Step Interval (int). Required parameter.

       time_step_interval_unit: Time Step Interval Unit (str). Required parameter.
          Choice list:['Years', 'Months', 'Weeks', 'Days', 'Hours', 'Minutes', 'Seconds', 'Milliseconds']

       time_step_alignment: Time Step Alignment (str). Optional parameter.
          Choice list:['EndTime', 'StartTime', 'ReferenceTime']

       time_step_reference: Time Step Reference (datetime). Optional parameter.

       summary_fields: Summary Fields (str). Optional parameter.

       output_name: Output Name (str). Required parameter.

       context: Context (str). Optional parameter.

        gis: Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


    Returns:
       output_cube - Output Space Time Cube as a DataFile

    """
    kwargs=locals()

    gis=_arcgis.env.active_gis if gis is None else gis
    url=gis.properties.helperServices.geoanalytics.url

    params={}
    for key, value in kwargs.items():
        if value is not None:
            params[key]=value

    _set_context(params)

    param_db={
        "point_layer": (_FeatureSet, "pointLayer"),
        "bin_size": (float, "binSize"),
        "bin_size_unit": (str, "binSizeUnit"),
        "time_step_interval": (int, "timeStepInterval"),
        "time_step_interval_unit": (str, "timeStepIntervalUnit"),
        "time_step_alignment": (str, "timeStepAlignment"),
        "time_step_reference": (_datetime, "timeStepReference"),
        "summary_fields": (str, "summaryFields"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output_cube": (DataFile, "Output Space Time Cube"),
    }
    return_values=[
        {"name": "output_cube", "display_name": "Output Space Time Cube", "type": DataFile},
    ]

    return _execute_gp_tool(gis, "CreateSpaceTimeCube", params, param_db, return_values, _use_async, url, True)




import json
import string
import random

import arcgis
from arcgis.gis import Layer
from arcgis.features import FeatureCollection

def _id_generator(size=6, chars=string.ascii_uppercase + string.digits):
    return ''.join(random.choice(chars) for _ in range(size))


def _feature_input(self, input_layer):
    input_layer_url = ""
    if isinstance(input_layer, arcgis.gis.Item):
        if input_layer.type.lower() == 'feature service':
            input_param =  {'url': input_layer.layers[0].url}
        elif input_layer.type.lower() == 'feature collection':
            fcdict = input_layer.get_data()
            fc = FeatureCollection(fcdict['layers'][0])
            input_param =  fc.layer
        else:
            raise TypeError("item type must be feature service or feature collection")

    elif isinstance(input_layer, arcgis.features.FeatureLayerCollection):
        input_layer_url = input_layer.layers[0].url
        input_param =  {"url": input_layer_url }

    elif isinstance(input_layer, FeatureCollection):
        input_param =  input_layer.properties

    elif isinstance(input_layer, Layer):
        input_layer_url = input_layer.url
        input_param =  {"url": input_layer_url }

    elif isinstance(input_layer, dict):
        input_param =  input_layer

    elif isinstance(input_layer, str):
        input_layer_url = input_layer
        input_param =  {"url": input_layer_url }

    else:
        raise Exception("Invalid format of input layer. url string, feature service Item, feature service instance or dict supported")

    return input_param


def _set_context(params):
    out_sr = arcgis.env.out_spatial_reference
    process_sr = arcgis.env.process_spatial_reference
    out_extent = arcgis.env.analysis_extent
    output_datastore = arcgis.env.output_datastore
    default_aggregation_styles = arcgis.env.default_aggregation_styles

    context = {}
    set_context = False

    if default_aggregation_styles is not None and\
       isinstance(default_aggregation_styles, bool):
        context['defaultAggregationStyles'] = default_aggregation_styles
        set_context = True
    if out_sr is not None:
        context['outSR'] = {'wkid': int(out_sr)}
        set_context = True
    if out_extent is not None:
        context['extent'] = out_extent
        set_context = True
    if process_sr is not None:
        context['processSR'] = {'wkid': int(process_sr)}
        set_context = True
    if output_datastore is not None:
        context['dataStore'] = output_datastore
        set_context = True

    if set_context:
        params["context"] = json.dumps(context)



def _create_output_service(gis, output_name, output_service_name='Analysis feature service', task='GeoAnalytics'):
    ok = gis.content.is_service_name_available(output_name, 'Feature Service')
    if not ok:
        raise RuntimeError("A Feature Service by this name already exists: " + output_name)

    createParameters = {
            "currentVersion": 10.2,
            "serviceDescription": "",
            "hasVersionedData": False,
            "supportsDisconnectedEditing": False,
            "hasStaticData": True,
            "maxRecordCount": 2000,
            "supportedQueryFormats": "JSON",
            "capabilities": "Query",
            "description": "",
            "copyrightText": "",
            "allowGeometryUpdates": False,
            "syncEnabled": False,
            "editorTrackingInfo": {
                "enableEditorTracking": False,
                "enableOwnershipAccessControl": False,
                "allowOthersToUpdate": True,
                "allowOthersToDelete": True
            },
            "xssPreventionInfo": {
                "xssPreventionEnabled": True,
                "xssPreventionRule": "InputOnly",
                "xssInputRule": "rejectInvalid"
            },
            "tables": [],
            "name": output_service_name.replace(' ', '_'),
            "options": {
                "dataSourceType": "spatiotemporal"
            }
        }

    output_service = gis.content.create_service(output_name, create_params=createParameters, service_type="featureService")
    description = "Feature Service generated from running the " + task + " tool."
    item_properties = {
            "description" : description,
            "tags" : "Analysis Result, " + task,
            "snippet": output_service_name
            }
    output_service.update(item_properties)
    return output_service

